/* product name: RXJIS                                                */
/* Version:      0.86                                                 */
/* author:       YANO Takashi                                         */
/* target:       OS/2 Warp J3.0+ (J2.1 is possible)                   */
/* module name:  RXJIS.DLL                                            */
/* source name:  RXJIS.CPP RXJIS.H                                    */
/* compiler:     VisualAge C++ 3.0                                    */
/* address:      tyano@ca2.so-net.or.jp or tyano@yamato.ibm.co.jp     */
/* comment:      RXJIS is a utility functions for REXX.               */
/*               It encodes and decodes JIS 7 bit strings and MIME    */
/*               format strings.                                      */
/*                                                                    */
/*               It provides the functions:                           */
/*               RxJisToPc                                            */
/*               RxJisJisTo                                           */
/*               RxJisDropFuncs                                       */
/*               RxJisToJis                                           */
/*               RxJisToBase64                                        */
/*               RxJisToQuotedPrintable                               */
/*               RxJisBase64To                                        */
/*               RxJisEucTo                                           */
/*               RxJisToEuc                                           */
/*               RxJisQuotedPrintableTo                               */
/*               RxJisLoadFuncs                                       */
/*               RxJisMimeJisTo                                       */
/*               RxJisJisCharJisTo                                    */
/*               RxJisJisToCharJisTo                                  */
/*               RxJisJisIso8859_1ToCp850                             */
/*               RxJisJisIso8859_2ToCp852                             */
/*               RxJisJisIso8859_9ToCp857                             */
/*               RxJisJisCp850ToIso8859_1                             */
/*               RxJisJisCp852ToIso8859_2                             */
/*               RxJisJisCp857ToIso8859_9                             */
/*               RxJisJisToZenkakuKatakana                            */
/*               RxJisJisWord                                         */
/*               RxJisJisWords                                        */
/*               RxJisJisWordIndex                                    */
/* history:      1996-2-5  0.00 initial release                       */
/*               1996-2-14 0.10 if find alone reset sequence,         */
/*                         add the set seq at the head.               */
/*                         reduce dll size                            */
/*               1996-3-20 0.20 JIS90 support                         */
/*                              Handle umatch JIS sequence            */
/*                              EUC support                           */
/*               1996-4-21 0.30 JIS78 support                         */
/*                              Performance inprovement               */
/*                              Add .HPP/.LIB                         */
/*               1996-5-1  0.40 Add RxJisMimeJis                      */
/*               1996-5-19 0.50 Add RxJisToDbcs                       */
/*                              Add RxJisToSbcs                       */
/*                              Add RxJisToAscii                      */
/*                              Add RxJisToUpper                      */
/*               1996-5-24 0.51 Fix RxJisToJis bug                    */
/*               1996-5-27 0.52 Fix RxJisToJis bug                    */
/*                              Fix Version No                        */
/*               1996-6-20 0.60 Fix RxJisToDbcs bug                   */
/*                              Try to use JIS instead of IBM unique  */
/*                              in RxJisToJis/RxJisToEuc              */
/*                              Add RxJisJisIso8859_1ToCp850          */
/*                              Add RxJisJisIso8859_2ToCp852          */
/*                              Add RxJisJisIso8859_9ToCp857          */
/*                              Add RxJisJisCp850ToIso8859_1          */
/*                              Add RxJisJisCp852ToIso8859_2          */
/*                              Add RxJisJisCp857ToIso8859_9          */
/*                              Fix RxJisQuotedPrintable bug          */
/*                              Fix JisXX NULL term bug               */
/*               1996-6-26 0.70 Add RxJisToZenkakuKatakana            */
/*                              Add JIS codeset 3 support             */
/*               1996-7-8  0.71 RxJisBase64To/RxJisToBase64 may not   */
/*                              work as expected.                     */
/*                              (from 0.70)                           */
/*               1996-7-21 0.80 Add RxJisAppend                       */
/*                              Add RxJisWord                         */
/*                              Add RxJisWords                        */
/*                              Add RxJisWordIndex                    */
/*                              Add RxJisUrltoVar                     */
/*               1996-7-22 0.81 Fix RxJisToJis bug - > 255 source     */
/*               1996-8-19 0.82 Fix RxJisAppend -> blanks + char      */
/*               1996-9-10 0.83 Add codepage 943 support              */
/*                              Fix Ver bug                           */
/*               1996-9-14 0.84 cp943 has NEC chars                   */
/*                                                                    */
/*                                                                    */
/*                                                                    */

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <itrace.hpp>
#define INCL_DOSMODULEMGR
#define INCL_DOSERRORS
#define INCL_DOSMISC 
#define INCL_DOSNLS
#include <os2.h>
#define INCL_RXFUNC
#define INCL_RXSHV
#include <rexxsaa.h>
#include "rxjis.h"

#define CallingValid 0
#define CallingInvalid 40
#define RxJisDll "RXJIS"
#define RxJisVersion "0.86" 
#define Base64Error -1
#define QuotedPrintableError -1
#define Base64Pad 64
#define Base64BitWidth 6
#define ByteBitWidth 8
#define MaxMimeLineSize 76
#define ByteMask 0x000000ff
#define Base64Mask 0x00003f
#define HighByteMask 0x0000ff00
#define Ht '\x09'
#define Space ' '
#define Esc '\x1b'
#define EucMsb 0x0080                     
#define ByteMsb 0x0080                     
#define EucMask 0x007f                    
#define EucSs2 0x008e                    
#define EucSs3 0x008f
#define JisEnvVarName "IPF_KEYS"          
#define Jis90Value "NEWJIS"               
#define Jis83InSeq DoubleByteSeqTable[0]  
#define Jis83OutSeq DoubleByteSeqTable[6] 
#define JisSpace 0x2121                   
#define RxJisError -1                     
#define SI 0x0e
#define SO 0x0f
#define CpListSize 8
#define RxJisArgCheck(source, target, targetsize, sourcesize) \
   if (NULL == source || NULL == target || targetsize <= 0) { \
      return (RxJisError); \
   } /* endif */ \
   if (sourcesize <= 0) { \
      sourcesize = strlen(source); \
   } /* endif */

#define RxJisCountCheck(count, targetsize) \
   if (count >= targetsize) { \
      return (RxJisError); \
   } /* endif */

#define RxJisCountCheck2(count, targetsize, sourcewide) \
   if (count >= targetsize) { \
      delete sourcewide; \
      return (RxJisError); \
   } /* endif */

#define SetValueInTable(t, l, h, f, c, v) \
   if (!t[h]) { \
      t[h] = new wchar_t[UCHAR_MAX+1]; \
      for (int c = 0; c <= UCHAR_MAX; c += 1) { \
         t[h][c] = f; \
      } /* endfor */ \                     
   } /* endif */ \
   t[h][l] = v;

#define IsDbcs1st(c) DbcsVec[c]

extern "C" {
RexxFunctionHandler RxJisLoadFuncs;
RexxFunctionHandler RxJisDropFuncs;
RexxFunctionHandler RxJisToJis;
RexxFunctionHandler RxJisJisTo; 
RexxFunctionHandler RxJisToEuc; 
RexxFunctionHandler RxJisToPc;
RexxFunctionHandler RxJisToBase64;
RexxFunctionHandler RxJisToQuotedPrintable;
RexxFunctionHandler RxJisBase64To;
RexxFunctionHandler RxJisEucTo; 
RexxFunctionHandler RxJisMimeJisTo; 
RexxFunctionHandler RxJisQuotedPrintableTo;
RexxFunctionHandler RxJisJisCharTo; 
RexxFunctionHandler RxJisToJisChar; 
RexxFunctionHandler RxJisToSbcs; 
RexxFunctionHandler RxJisToDbcs; 
RexxFunctionHandler RxJisToAscii; 
RexxFunctionHandler RxJisToUpper; 
RexxFunctionHandler RxJisToZenkakuKatakana; 
RexxFunctionHandler RxJisIso8859_1ToCp850; 
RexxFunctionHandler RxJisIso8859_2ToCp852; 
RexxFunctionHandler RxJisIso8859_9ToCp857; 
RexxFunctionHandler RxJisCp850ToIso8859_1; 
RexxFunctionHandler RxJisCp852ToIso8859_2; 
RexxFunctionHandler RxJisCp857ToIso8859_9; 
RexxFunctionHandler RxJisWords; 
RexxFunctionHandler RxJisWord; 
RexxFunctionHandler RxJisWordIndex;        
RexxFunctionHandler RxJisAppend;        
RexxFunctionHandler RxJisUrlToVar;        
};

static PSZ RxJisFuncsTable[] = {
   "RxJisLoadFuncs",
   "RxJisDropFuncs",
   "RxJisToBase64",
   "RxJisToQuotedPrintable",
   "RxJisToJis",
   "RxJisToEuc", 
   "RxJisToPc",
   "RxJisJisTo", 
   "RxJisBase64To",
   "RxJisEucTo", 
   "RxJisMimeJisTo", 
   "RxJisQuotedPrintableTo",
   "RxJisJisCharTo", 
   "RxJisToJisChar", 
   "RxJisToSbcs", 
   "RxJisToDbcs", 
   "RxJisToAscii", 
   "RxJisToUpper", 
   "RxJisToZenkakuKatakana", 
   "RxJisIso8859_1ToCp850", 
   "RxJisIso8859_2ToCp852", 
   "RxJisIso8859_9ToCp857", 
   "RxJisCp850ToIso8859_1", 
   "RxJisCp852ToIso8859_2", 
   "RxJisCp857ToIso8859_9", 
   "RxJisWords", 
   "RxJisWord", 
   "RxJisWordIndex", 
   "RxJisAppend", 
   "RxJisUrlToVar", 
};

static const size_t NoOfEntries = sizeof(RxJisFuncsTable) / sizeof(RxJisFuncsTable[0]);

typedef wchar_t *PWCHAR_T;
static PWCHAR_T Jis78ToPccodeTable[UCHAR_MAX+1];
static PWCHAR_T Jis83ToPccodeTable[UCHAR_MAX+1];
static PWCHAR_T Jis90ToPccodeTable[UCHAR_MAX+1];
static PWCHAR_T JisX0212_1990ToPccodeTable[UCHAR_MAX+1];
static PWCHAR_T PccodeToJis83Table[UCHAR_MAX+1];
static PWCHAR_T UppercaseTable[UCHAR_MAX+1];
static PCHAR *ToSbcsTable[UCHAR_MAX+1];
static BOOL TableIsLoaded = FALSE;
static APIRET LoadRxJisTables(void);
static BOOL (APIENTRY16 *NlsFindWord)(USHORT codepage, PSZ string, ULONG strlen, ULONG  position, PULONG pStartWord, PULONG pEndWord, PULONG pNextWord);
static USHORT DefaultCodepage;
static int Base64To(PCHAR target, PCHAR source, int targetsize, int sourcesize = 0, int linesize = 0);          
static int QuotedPrintableTo( PCHAR target, PCHAR source, int targetsize, int sourcesize = 0, int linesize = 0); 
static int ToPc(PCHAR target, PCHAR source, int targetsize, int sourcesize = 0, int linesize = 0); 
static wchar_t PcCodeToJisCode(wchar_t c); 

static wchar_t ToDbcsTable1[UCHAR_MAX+1];
static wchar_t ToDbcsTable2[UCHAR_MAX+1];
static wchar_t ToDbcsTable3[UCHAR_MAX+1];
static PWCHAR_T ToDbcsTable4[UCHAR_MAX+1];
static PWCHAR_T ToDbcsTable5[UCHAR_MAX+1];
#define ToIso8859_1(c) Cp850ToIso8859_1Table[c]
#define ToCp850(c) Iso8859_1ToCp850Table[c]
static CHAR Cp850ToIso8859_1Table[UCHAR_MAX+1];
static CHAR Iso8859_1ToCp850Table[UCHAR_MAX+1] = {
   0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
   0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
   0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
   0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F,
   0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F,
   0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F,
   0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x6B, 0x6C, 0x6D, 0x6E, 0x6F,
   0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A, 0x7B, 0x7C, 0x7D, 0x7E, 0x7F,
   0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8A, 0x8B, 0x8C, 0x8D, 0x8E, 0x8F,
   0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9A, 0x9B, 0x9C, 0x9D, 0x9E, 0x9F,
   0xFF, 0xAD, 0xBD, 0x9C, 0xCF, 0xBE, 0xDD, 0xF5, 0xF9, 0xB8, 0xA6, 0xAE, 0xAA, 0xF0, 0xA9, 0xEE,
   0xF8, 0xF1, 0xFD, 0xFC, 0xEF, 0xE6, 0xF4, 0xFA, 0xF7, 0xFB, 0xA7, 0xAF, 0xAC, 0xAB, 0xF3, 0xA8,
   0xB7, 0xB5, 0xB6, 0xC7, 0x8E, 0x8F, 0x92, 0x80, 0xD4, 0x90, 0xD2, 0xD3, 0xDE, 0xD6, 0xD7, 0xD8,
   0xD1, 0xA5, 0xE3, 0xE0, 0xE2, 0xE5, 0x99, 0x9E, 0x9D, 0xEB, 0xE9, 0xEA, 0x9A, 0xED, 0xE8, 0xE1,
   0x85, 0xA0, 0x83, 0xC6, 0x84, 0x86, 0x91, 0x87, 0x8A, 0x82, 0x88, 0x89, 0x8D, 0xA1, 0x8C, 0x8B,
   0xD0, 0xA4, 0x95, 0xA2, 0x93, 0xE4, 0x94, 0xF6, 0x9B, 0x97, 0xA3, 0x96, 0x81, 0xEC, 0xE7, 0x98,
};

#define ToIso8859_2(c) Cp852ToIso8859_2Table[c]
#define ToCp852(c) Iso8859_2ToCp852Table[c]
static CHAR Cp852ToIso8859_2Table[UCHAR_MAX+1];
static CHAR Iso8859_2ToCp852Table[UCHAR_MAX+1] = {
   0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
   0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
   0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
   0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F,
   0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F,
   0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F,
   0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x6B, 0x6C, 0x6D, 0x6E, 0x6F,
   0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A, 0x7B, 0x7C, 0x7D, 0x7E, 0x7F,
   0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8A, 0x8B, 0x8C, 0x8D, 0x8E, 0x8F,
   0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9A, 0x9B, 0x9C, 0x9D, 0x9E, 0x9F,
   0xFF, 0xA4, 0xF4, 0x9D, 0xCF, 0x95, 0x97, 0xF5, 0xF9, 0xE6, 0xB8, 0x9B, 0x8D, 0xF0, 0xA6, 0xBD,
   0xF8, 0xA5, 0xF2, 0x88, 0xEF, 0x96, 0x98, 0xF3, 0xF7, 0xE7, 0xAD, 0x9C, 0xAB, 0xF1, 0xA7, 0xBE,
   0xE8, 0xB5, 0xB6, 0xC6, 0x8E, 0x91, 0x8F, 0x80, 0xAC, 0x90, 0xA8, 0xD3, 0xB7, 0xD6, 0xD7, 0xD2,
   0xD1, 0xE3, 0xD5, 0xE0, 0xE2, 0x8A, 0x99, 0x9E, 0xFC, 0xDE, 0xE9, 0xEB, 0x9A, 0xED, 0xDD, 0xE1,
   0xEA, 0xA0, 0x83, 0xC6, 0x84, 0x92, 0x86, 0x87, 0x9F, 0x82, 0xA8, 0x89, 0xD8, 0xA1, 0x8C, 0xD4,
   0xD0, 0xE4, 0xE5, 0xA2, 0x93, 0x8B, 0x94, 0xF6, 0xFD, 0x85, 0xA3, 0xFB, 0x81, 0xEC, 0xEE, 0xFA,
};

#define ToIso8859_9(c) Cp857ToIso8859_9Table[c]
#define ToCp857(c) Iso8859_9ToCp857Table[c]
static CHAR Cp857ToIso8859_9Table[UCHAR_MAX+1];
static CHAR Iso8859_9ToCp857Table[UCHAR_MAX+1] = {
   0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
   0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
   0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
   0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F,
   0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F,
   0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F,
   0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x6B, 0x6C, 0x6D, 0x6E, 0x6F,
   0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A, 0x7B, 0x7C, 0x7D, 0x7E, 0x7F,
   0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8A, 0x8B, 0x8C, 0x8D, 0x8E, 0x8F,
   0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9A, 0x9B, 0x9C, 0x9D, 0x9E, 0x9F,
   0xFF, 0xAD, 0xBD, 0x9C, 0xCF, 0xBE, 0xDD, 0xF5, 0xF9, 0xB8, 0xD1, 0xAE, 0xAA, 0x1A, 0xA9, 0xEE,
   0xF8, 0xF1, 0xFD, 0xFC, 0xEF, 0xE6, 0xF4, 0xFA, 0xF7, 0xFB, 0xD0, 0xAF, 0xAC, 0xAB, 0xF3, 0xA8,
   0xB7, 0xB5, 0xB6, 0xC7, 0x8E, 0x8F, 0x92, 0x80, 0xD4, 0x90, 0xD2, 0xD3, 0xDE, 0xD6, 0xD7, 0xD8,
   0xA6, 0xA5, 0xE3, 0xE0, 0xE2, 0xE5, 0x99, 0xE8, 0x9D, 0xEB, 0xE9, 0xEA, 0x9A, 0x1A, 0x9E, 0xE1,
   0x85, 0xA0, 0x83, 0xC6, 0x84, 0x86, 0x91, 0x87, 0x8A, 0x82, 0x88, 0x89, 0xEC, 0xA1, 0x8C, 0x8B,
   0xA7, 0xA4, 0x95, 0xA2, 0x93, 0xE4, 0x94, 0xF6, 0x9B, 0x97, 0xA3, 0x96, 0x81, 0x8D, 0x9F, 0xED,
};

struct DbcsConvTable {
   wchar_t code1; /* IBM unique PC codes */
   wchar_t code2; /* Alternative PC codes in JIS */
   wchar_t code3; /* JIS level 3 kuten codes */
};

DbcsConvTable NecCharTable[] = {
   {0x2D21,  0x88EA}, // {0x2D21, L''}, /* 13 1 */
   {0x2D22,  0x836A}, // {0x2D22, L'j'}, /* 13 2 */
   {0x2D23,  0x8E4F}, // {0x2D23, L'O'}, /* 13 3 */
   {0x2D24,  0x8E6C}, // {0x2D24, L'l'}, /* 13 4 */
   {0x2D25,  0x8CDC}, // {0x2D25, L''}, /* 13 5 */
   {0x2D26,  0x985A}, // {0x2D26, L'Z'}, /* 13 6 */
   {0x2D27,  0x8EB5}, // {0x2D27, L''}, /* 13 7 */
   {0x2D28,  0x94AA}, // {0x2D28, L''}, /* 13 8 */
   {0x2D29,  0x8BE3}, // {0x2D29, L''}, /* 13 9 */
   {0x2D2A,  0x8F5C}, // {0x2D2A, L'\'}, /* 13 10 */
   {0x2D35,  0x88EA}, // {0x2D35, L''}, /* 13 21 */
   {0x2D36,  0x836A}, // {0x2D36, L'j'}, /* 13 22 */
   {0x2D37,  0x8E4F}, // {0x2D37, L'O'}, /* 13 23 */
   {0x2D38,  0x8E6C}, // {0x2D38, L'l'}, /* 13 24 */
   {0x2D39,  0x8CDC}, // {0x2D39, L''}, /* 13 25 */
   {0x2D3A,  0x985A}, // {0x2D3A, L'Z'}, /* 13 26 */
   {0x2D3B,  0x8EB5}, // {0x2D3B, L''}, /* 13 27 */
   {0x2D3C,  0x94AA}, // {0x2D3C, L''}, /* 13 28 */
   {0x2D3D,  0x8BE3}, // {0x2D3D, L''}, /* 13 29 */
   {0x2D3E,  0x8F5C}, // {0x2D3E, L'\'}, /* 13 30 */
   {0x2D40,  0x828D}, // {0x2D40, L''}, /* 13 32 */
   {0x2D41,  0x828B}, // {0x2D41, L''}, /* 13 33 */
   {0x2D42,  0x9157}, // {0x2D42, L'W'}, /* 13 34 */
   {0x2D43,  0x828D}, // {0x2D43, L''}, /* 13 35 */
   {0x2D44,  0x8287}, // {0x2D44, L''}, /* 13 36 */
   {0x2D45,  0x8294}, // {0x2D45, L''}, /* 13 37 */
   {0x2D46,  0x8281}, // {0x2D46, L''}, /* 13 38 */
   {0x2D48,  0x828C}, // {0x2D48, L''}, /* 13 40 */
   {0x2D49,  0x8276}, // {0x2D49, L'v'}, /* 13 41 */
   {0x2D4B,  0x8190}, // {0x2D4B, L''}, /* 13 43 */
   {0x2D4C,  0x8283}, // {0x2D4C, L''}, /* 13 44 */
   {0x2D4D,  0x8193}, // {0x2D4D, L''}, /* 13 45 */
   {0x2D4F,  0x95C5}, // {0x2D4F, L''}, /* 13 47 */
   {0x2D50,  0x96AE}, // {0x2D50, L''}, /* 13 48 */
   {0x2D51,  0x9157}, // {0x2D51, L'W'}, /* 13 49 */
   {0x2D52,  0x8BCC}, // {0x2D52, L''}, /* 13 50 */
   {0x2D53,  0xE147}, // {0x2D53, L'G'}, /* 13 51 */
   {0x2D54,  0xE143}, // {0x2D54, L'C'}, /* 13 52 */
   {0x2D55,  0xE28A}, // {0x2D55, L''}, /* 13 53 */
   {0x2D5F,  0x95BD}, // {0x2D5F, L''}, /* 13 63 */
   {0x2D60,  0x8167}, // {0x2D60, L'g'}, /* 13 64 */
   {0x2D61,  0x8168}, // {0x2D61, L'h'}, /* 13 65 */
   {0x2D62,  0x8194}, // {0x2D62, L''}, /* 13 66 */
   {0x2D64,  0x9364}, // {0x2D64, L'd'}, /* 13 68 */
   {0x2D65,  0x8FE3}, // {0x2D65, L''}, /* 13 69 */
   {0x2D66,  0x9286}, // {0x2D66, L''}, /* 13 70 */
   {0x2D67,  0x89BA}, // {0x2D67, L''}, /* 13 71 */
   {0x2D68,  0x8DB6}, // {0x2D68, L''}, /* 13 72 */
   {0x2D69,  0x8945}, // {0x2D69, L'E'}, /* 13 73 */
   {0x2D6A,  0x8A94}, // {0x2D6A, L''}, /* 13 74 */
   {0x2D6B,  0x974C}, // {0x2D6B, L'L'}, /* 13 75 */
   {0x2D6C,  0x91E3}, // {0x2D6C, L''}, /* 13 76 */
   {0x2D6D,  0x96BE}, // {0x2D6D, L''}, /* 13 77 */
   {0x2D6E,  0x91E5}, // {0x2D6E, L''}, /* 13 78 */
   {0x2D6F,  0x8FBA}, // {0x2D6F, L''}, /* 13 79 */
   {0x2D70,  0x81E0}, // {0x2D70, L''}, /* 13 80 */
   {0x2D71,  0x81DF}, // {0x2D71, L''}, /* 13 81 */
   {0x2D72,  0x81E7}, // {0x2D72, L''}, /* 13 82 */
   {0x2D73,  0x83B0}, // {0x2D73, L''}, /* 13 83 */
   {0x2D74,  0x81E3}, // {0x2D74, L''}, /* 13 84 */
   {0x2D75,  0x81DB}, // {0x2D75, L''}, /* 13 85 */
   {0x2D76,  0x81DA}, // {0x2D76, L''}, /* 13 86 */
   {0x2D79,  0x81E6}, // {0x2D79, L''}, /* 13 89 */
   {0x2D7A,  0x81BF}, // {0x2D7A, L''}, /* 13 90 */
   {0x2D7B,  0x81BE}, // {0x2D7B, L''}, /* 13 91 */
   {0x7C71,  0x88EA}, // {0x7C71, L''}, /* 92 81 */
   {0x7C72,  0x836A}, // {0x7C72, L'j'}, /* 92 82 */
   {0x7C73,  0x8E4F}, // {0x7C73, L'O'}, /* 92 83 */
   {0x7C74,  0x8E6C}, // {0x7C74, L'l'}, /* 92 84 */
   {0x7C75,  0x8CDC}, // {0x7C75, L''}, /* 92 85 */
   {0x7C76,  0x985A}, // {0x7C76, L'Z'}, /* 92 86 */
   {0x7C77,  0x8EB5}, // {0x7C77, L''}, /* 92 87 */
   {0x7C78,  0x94AA}, // {0x7C78, L''}, /* 92 88 */
   {0x7C79,  0x8BE3}, // {0x7C79, L''}, /* 92 89 */
   {0x7C7A,  0x8F5C}, // {0x7C7A, L'\'}, /* 92 90 */
   {0x7C7B,  0x81CA}, // {0x7C7B, L''}, /* 92 91 */
   {0x7C7C,  0x8162}, // {0x7C7C, L'b'}, /* 92 92 */
   {0x7C7D,  0x8165}, // {0x7C7D, L'e'}, /* 92 93 */
   {0x7C7E,  0x8167}, // {0x7C7E, L'g'}, /* 92 94 */
};

DbcsConvTable IbmCharTable[] = {
   {0xFA40, 0x88EA, 0000}, // {L'@', L'', 0000},
   {0xFA41, 0x93F1, 0000}, // {L'A', L'', 0000},
   {0xFA42, 0x8E4F, 0000}, // {L'B', L'O', 0000},
   {0xFA43, 0x8E6C, 0000}, // {L'C', L'l', 0000},
   {0xFA44, 0x8CDC, 0000}, // {L'D', L'', 0000},
   {0xFA45, 0x985A, 0000}, // {L'E', L'Z', 0000},
   {0xFA46, 0x8EB5, 0000}, // {L'F', L'', 0000},
   {0xFA47, 0x94AA, 0000}, // {L'G', L'', 0000},
   {0xFA48, 0x8BE3, 0000}, // {L'H', L'', 0000},
   {0xFA49, 0x8F5C, 0000}, // {L'I', L'\', 0000},
   {0xFA4A, 0x88EA, 0000}, // {L'J', L'', 0000},
   {0xFA4B, 0x93F1, 0000}, // {L'K', L'', 0000},
   {0xFA4C, 0x8E4F, 0000}, // {L'L', L'O', 0000},
   {0xFA4D, 0x8E6C, 0000}, // {L'M', L'l', 0000},
   {0xFA4E, 0x8CDC, 0000}, // {L'N', L'', 0000},
   {0xFA4F, 0x985A, 0000}, // {L'O', L'Z', 0000},
   {0xFA50, 0x8EB5, 0000}, // {L'P', L'', 0000},
   {0xFA51, 0x94AA, 0000}, // {L'Q', L'', 0000},
   {0xFA52, 0x8BE3, 0000}, // {L'R', L'', 0000},
   {0xFA53, 0x8F5C, 0000}, // {L'S', L'\', 0000},
   {0xFA54, 0x81CA, 0000}, // {L'T', L'', 0000},
   {0xFA55, 0x8162, 0235}, // {L'U', L'b', 0235},
   {0xFA56, 0x8165, 0000}, // {L'V', L'e', 0000},
   {0xFA57, 0x8167, 0000}, // {L'W', L'g', 0000},
   {0xFA58, 0x8A94, 0000}, // {L'X', L'', 0000},
   {0xFA59, 0x81F2, 0000}, // {L'Y', L'', 0000},
   {0xFA5A, 0x9364, 0000}, // {L'Z', L'd', 0000},
   {0xFA5B, 0x81E6, 0000}, // {L'[', L'', 0000},
   {0xFA5C, 0xE345, 5267}, // {L'\', L'E', 5267},
   {0xFA5D, 0x9645, 6063}, // {L']', L'E', 6063},
   {0xFA5E, 0xFA5E, 6873}, // {L'^', L'^', 6873}, /*?*/
   {0xFA5F, 0xFA5F, 6788}, // {L'_', L'_', 6788}, /*?*/
   {0xFA60, 0xFA60, 5701}, // {L'`', L'`', 5701}, /*?*/
   {0xFA61, 0xFA61, 0000}, // {L'a', L'a', 0000}, /*?*/
   {0xFA62, 0xFA62, 1727}, // {L'b', L'b', 1727}, /*?*/
   {0xFA63, 0xFA63, 3413}, // {L'c', L'c', 3413}, /*?*/
   {0xFA64, 0xFA64, 3592}, // {L'd', L'd', 3592}, /*?*/
   {0xFA65, 0xFA65, 6848}, // {L'e', L'e', 6848}, /*?*/
   {0xFA66, 0xFA66, 3431}, // {L'f', L'f', 3431}, /*?*/
   {0xFA67, 0xFA67, 2884}, // {L'g', L'g', 2884}, /*?*/
   {0xFA68, 0x84A0, 1609}, // {L'h', L'', 1609},
   {0xFA69, 0xFA69, 1640}, // {L'i', L'i', 1640},
   {0xFA6A, 0xFA6A, 0000}, // {L'j', L'j', 0000}, /*?*/
   {0xFA6B, 0xFA6B, 1650}, // {L'k', L'k', 1650}, /*?*/
   {0xFA6C, 0xFA6C, 1652}, // {L'l', L'l', 1652}, /*?*/
   {0xFA6D, 0xFA6D, 1667}, // {L'm', L'm', 1667}, /*?*/
   {0xFA6E, 0xFA6E, 1678}, // {L'n', L'n', 1678}, /*?*/
   {0xFA6F, 0xFA6F, 1707}, // {L'o', L'o', 1707}, /*?*/
   {0xFA70, 0xFA70, 1703}, // {L'p', L'p', 1703}, /*?*/
   {0xFA71, 0x9C6D, 1712}, // {L'q', L'm', 1712},
   {0xFA72, 0xFA72, 1709}, // {L'r', L'r', 1709},
   {0xFA73, 0xFA73, 1730}, // {L's', L's', 1730}, /*?*/
   {0xFA74, 0x8970, 1763}, // {L't', L'p', 1763},
   {0xFA75, 0xFA75, 1756}, // {L'u', L'u', 1756},
   {0xFA76, 0xFA76, 1740}, // {L'v', L'v', 1740}, /*?*/
   {0xFA77, 0xFA77, 1755}, // {L'w', L'w', 1755}, /*?*/
   {0xFA78, 0xFA78, 1767}, // {L'x', L'x', 1767}, /*?*/
   {0xFA79, 0xFA79, 1784}, // {L'y', L'y', 1784}, /*?*/
   {0xFA7A, 0xFA7A, 1765}, // {L'z', L'z', 1765}, /*?*/
   {0xFA7B, 0xFA7B, 1803}, // {L'{', L'{', 1803},
   {0xFA7C, 0xFA7C, 0000}, // {L'|', L'|', 0000},
   {0xFA7D, 0xFA7D, 1827}, // {L'}', L'}', 1827}, /*?*/
   {0xFA7E, 0x995B, 1870}, // {L'~', L'[', 1870},
   {0xFA80, 0xFA80, 1877}, // {L'', L'', 1877}, /*?*/
   {0xFA81, 0xFA81, 1885}, // {L'', L'', 1885}, /*?*/
   {0xFA82, 0xFA82, 1892}, // {L'', L'', 1892}, /*?*/
   {0xFA83, 0x9597, 0000}, // {L'', L'', 0000},
   {0xFA84, 0xFA84, 1921}, // {L'', L'', 1921}, /*?*/
   {0xFA85, 0xFA85, 1956}, // {L'', L'', 1956}, /*?*/
   {0xFA86, 0xFA86, 1959}, // {L'', L'', 1959}, /*?*/
   {0xFA87, 0xFA87, 1969}, // {L'', L'', 1969}, /*?*/
   {0xFA88, 0xFA88, 1978}, // {L'', L'', 1978}, /*?*/
   {0xFA89, 0xFA89, 1991}, // {L'', L'', 1991},
   {0xFA8A, 0x99B0, 0000}, // {L'', L'', 0000},
   {0xFA8B, 0xFA8B, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFA8C, 0xFA8C, 2032}, // {L'', L'', 2032}, /*?*/
   {0xFA8D, 0xFA8D, 2039}, // {L'', L'', 2039}, /*?*/
   {0xFA8E, 0xFA8E, 2048}, // {L'', L'', 2048},
   {0xFA8F, 0xFA8F, 2062}, // {L'', L'', 2062}, /*?*/
   {0xFA90, 0xFA90, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFA91, 0x9A42, 2110}, // {L'', L'B', 2110},
   {0xFA92, 0x9861, 0000}, // {L'', L'a', 0000},
   {0xFA93, 0xFA93, 2115}, // {L'', L'', 2115}, /*?*/
   {0xFA94, 0xFA94, 2136}, // {L'', L'', 2136},
   {0xFA95, 0x934E, 2172}, // {L'', L'N', 2172},
   {0xFA96, 0xFA96, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFA97, 0xFA97, 2334}, // {L'', L'', 2334}, /*?*/
   {0xFA98, 0xFA98, 2368}, // {L'', L'', 2368}, /*?*/
   {0xFA99, 0xFA99, 2372}, // {L'', L'', 2372}, /*?*/
   {0xFA9A, 0xFA9A, 2371}, // {L'', L'', 2371}, /*?*/
   {0xFA9B, 0xFA9B, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFA9C, 0x92CB, 0000}, // {L'', L'', 0000},
   {0xFA9D, 0x919D, 0000}, // {L'', L'', 0000},
   {0xFA9E, 0xFA9E, 2446}, // {L'', L'', 2446}, /*?*/
   {0xFA9F, 0xFA9F, 2465}, // {L'', L'', 2465}, /*?*/
   {0xFAA0, 0xFAA0, 2485}, // {L'', L'', 2485}, /*?*/
   {0xFAA1, 0xFAA1, 2487}, // {L'', L'', 2487}, /*?*/
   {0xFAA2, 0xFAA2, 2488}, // {L'', L'', 2488}, /*?*/
   {0xFAA3, 0xFAA3, 2492}, // {L'', L'', 2492}, /*?*/
   {0xFAA4, 0xFAA4, 2515}, // {L'', L'', 2515}, /*?*/
   {0xFAA5, 0xFAA5, 2523}, // {L'', L'', 2523}, /*?*/
   {0xFAA6, 0xFAA6, 2630}, // {L'', L'', 2630}, /*?*/
   {0xFAA7, 0xFAA7, 2659}, // {L'', L'', 2659}, /*?*/
   {0xFAA8, 0x944A, 4510}, // {L'', L'J', 4510},
   {0xFAA9, 0xFAA9, 2665}, // {L'', L'', 2665},
   {0xFAAA, 0x8AB0, 0000}, // {L'', L'', 0000},
   {0xFAAB, 0xE099, 2675}, // {L'', L'', 2675},
   {0xFAAC, 0xFAAC, 2719}, // {L'', L'', 2719}, /*?*/
   {0xFAAD, 0xFAAD, 2724}, // {L'', L'', 2724}, /*?*/
   {0xFAAE, 0xFAAE, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFAAF, 0x9093, 2742}, // {L'', L'', 2742},
   {0xFAB0, 0x9BC7, 0000}, // {L'', L'', 0000},
   {0xFAB1, 0x8DE8, 0000}, // {L'', L'', 0000},
   {0xFAB2, 0xFAB2, 2748}, // {L'', L'', 2748}, /*?*/
   {0xFAB3, 0xFAB3, 2762}, // {L'', L'', 2762}, /*?*/
   {0xFAB4, 0xFAB4, 2784}, // {L'', L'', 2784},
   {0xFAB5, 0x9388, 2785}, // {L'', L'', 2785},
   {0xFAB6, 0xFAB6, 2789}, // {L'', L'', 2789}, /*?*/
   {0xFAB7, 0xFAB7, 2868}, // {L'', L'', 2868}, /*?*/
   {0xFAB8, 0xFAB8, 2877}, // {L'', L'', 2877}, /*?*/
   {0xFAB9, 0xFAB9, 2894}, // {L'', L'', 2894}, /*?*/
   {0xFABA, 0x93BF, 0000}, // {L'', L'', 0000},
   {0xFABB, 0xFABB, 2934}, // {L'', L'', 2934},
   {0xFABC, 0xFABC, 2971}, // {L'', L'', 2971}, /*?*/
   {0xFABD, 0x8978, 0000}, // {L'', L'x', 0000},
   {0xFABE, 0x934E, 2980}, // {L'', L'N', 2980}, /*?*/
   {0xFABF, 0xFABF, 3016}, // {L'', L'', 3016}, /*?*/
   {0xFAC0, 0xFAC0, 3012}, // {L'', L'', 3012},
   {0xFAC1, 0x9CB0, 0000}, // {L'', L'', 0000},
   {0xFAC2, 0xFAC2, 3019}, // {L'', L'', 3019}, /*?*/
   {0xFAC3, 0xFAC3, 3029}, // {L'', L'', 3029}, /*?*/
   {0xFAC4, 0xFAC4, 3045}, // {L'', L'', 3045}, /*?*/
   {0xFAC5, 0xFAC5, 3041}, // {L'', L'', 3041}, /*?*/
   {0xFAC6, 0x8AEC, 3068}, // {L'', L'', 3068},
   {0xFAC7, 0xFAC7, 3108}, // {L'', L'', 3108}, /*?*/
   {0xFAC8, 0xFAC8, 3141}, // {L'', L'', 3141}, /*?*/
   {0xFAC9, 0xFAC9, 3236}, // {L'', L'', 3236}, /*?*/
   {0xFACA, 0xFACA, 3268}, // {L'', L'', 3268}, /*?*/
   {0xFACB, 0xFACB, 3284}, // {L'', L'', 3284}, /*?*/
   {0xFACC, 0xFACC, 3306}, // {L'', L'', 3306},
   {0xFACD, 0x8BB3, 0000}, // {L'', L'', 0000},
   {0xFACE, 0xFACE, 3385}, // {L'', L'', 3385}, /*?*/
   {0xFACF, 0xFACF, 3392}, // {L'', L'', 3392},
   {0xFAD0, 0x8D56, 3388}, // {L'', L'V', 3388},
   {0xFAD1, 0xFAD1, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFAD2, 0xFAD2, 3411}, // {L'', L'', 3411}, /*?*/
   {0xFAD3, 0xFAD3, 3401}, // {L'', L'', 3401}, /*?*/
   {0xFAD4, 0xFAD4, 3405}, // {L'', L'', 3405}, /*?*/
   {0xFAD5, 0xE1A8, 0000}, // {L'', L'', 0000},
   {0xFAD6, 0xFAD6, 3424}, // {L'', L'', 3424}, /*?*/
   {0xFAD7, 0xFAD7, 3426}, // {L'', L'', 3426}, /*?*/
   {0xFAD8, 0x90B0, 0000}, // {L'', L'', 0000},
   {0xFAD9, 0x9DF0, 3436}, // {L'', L'', 3436},
   {0xFADA, 0xFADA, 3450}, // {L'', L'', 3450}, /*?*/
   {0xFADB, 0xFADB, 3455}, // {L'', L'', 3455},
   {0xFADC, 0xFADC, 3459}, // {L'', L'', 3459}, /*?*/
   {0xFADD, 0xFADD, 3462}, // {L'', L'', 3462}, /*?*/
   {0xFADE, 0xFADE, 3477}, // {L'', L'', 3477}, /*?*/
   {0xFADF, 0xFADF, 3480}, // {L'', L'', 3480}, /*?*/
   {0xFAE0, 0x984E, 0000}, // {L'', L'N', 0000},
   {0xFAE1, 0xFAE1, 3501}, // {L'', L'', 3501}, /*?*/
   {0xFAE2, 0xFAE2, 3521}, // {L'', L'', 3521}, /*?*/
   {0xFAE3, 0x8C4B, 3541}, // {L'', L'K', 3541},
   {0xFAE4, 0xFAE4, 3525}, // {L'', L'', 3525}, /*?*/
   {0xFAE5, 0x96F6, 0000}, // {L'', L'', 0000},
   {0xFAE6, 0xFAE6, 3556}, // {L'', L'', 3556}, /*?*/
   {0xFAE7, 0xFAE7, 3594}, // {L'', L'', 3594}, /*?*/
   {0xFAE8, 0xFAE8, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFAE9, 0xFAE9, 3644}, // {L'', L'', 3644},
   {0xFAEA, 0x9F4F, 0000}, // {L'', L'O', 0000},
   {0xFAEB, 0x8BE9, 3657}, // {L'', L'', 3657},
   {0xFAEC, 0xFAEC, 3674}, // {L'', L'', 3674}, /*?*/
   {0xFAED, 0xFAED, 3693}, // {L'', L'', 3693}, /*?*/
   {0xFAEE, 0x89A1, 0000}, // {L'', L'', 0000},
   {0xFAEF, 0xFAEF, 3707}, // {L'', L'', 3707}, /*?*/
   {0xFAF0, 0xFAF0, 3721}, // {L'', L'', 3721}, /*?*/
   {0xFAF1, 0xFAF1, 3722}, // {L'', L'', 3722}, /*?*/
   {0xFAF2, 0xFAF2, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFAF3, 0xFAF3, 3753}, // {L'', L'', 3753}, /*?*/
   {0xFAF4, 0xFAF4, 3824}, // {L'', L'', 3824}, /*?*/
   {0xFAF5, 0xFAF5, 3855}, // {L'', L'', 3855},
   {0xFAF6, 0xFAF6, 3864}, // {L'', L'', 3864}, /*?*/
   {0xFAF7, 0xFAF7, 3874}, // {L'', L'', 3874},
   {0xFAF8, 0xFAF8, 3867}, // {L'', L'', 3867}, /*?*/
   {0xFAF9, 0xFAF9, 3901}, // {L'', L'', 3901}, /*?*/
   {0xFAFA, 0xFAFA, 3911}, // {L'', L'', 3911},
   {0xFAFB, 0xFAFB, 3939}, // {L'', L'', 3939},
   {0xFAFC, 0xFAFC, 3935}, // {L'', L'', 3935}, /*?*/
   {0xFB40, 0xE4AC, 3943}, // {L'@', L'', 3943},
   {0xFB41, 0xFB41, 3947}, // {L'A', L'A', 3947}, /*?*/
   {0xFB42, 0xFB42, 3957}, // {L'B', L'B', 3957}, /*?*/
   {0xFB43, 0x90B4, 0000}, // {L'C', L'', 0000},
   {0xFB44, 0xFB44, 0000}, // {L'D', L'D', 0000}, /*?*/
   {0xFB45, 0xFB45, 3970}, // {L'E', L'E', 3970},
   {0xFB46, 0xFB46, 3978}, // {L'F', L'F', 3978}, /*?*/
   {0xFB47, 0xFB47, 3992}, // {L'G', L'G', 3992}, /*?*/
   {0xFB48, 0xFB48, 3975}, // {L'H', L'H', 3975}, /*?*/
   {0xFB49, 0xFB49, 3980}, // {L'I', L'I', 3980}, /*?*/
   {0xFB4A, 0xFB4A, 4017}, // {L'J', L'J', 4017}, /*?*/
   {0xFB4B, 0xFB4B, 4069}, // {L'K', L'K', 4069}, /*?*/
   {0xFB4C, 0xFB4C, 4088}, // {L'L', L'L', 4088}, /*?*/
   {0xFB4D, 0x956C, 4106}, // {L'M', L'l', 4106},
   {0xFB4E, 0xFB4E, 4111}, // {L'N', L'N', 4111}, /*?*/
   {0xFB4F, 0xFB4F, 4113}, // {L'O', L'O', 4113}, /*?*/
   {0xFB50, 0x90A3, 0000}, // {L'P', L'', 0000},
   {0xFB51, 0xFB51, 4142}, // {L'Q', L'Q', 4142}, /*?*/
   {0xFB52, 0xFB52, 4151}, // {L'R', L'R', 4151},
   {0xFB53, 0xFB53, 4173}, // {L'S', L'S', 4173}, /*?*/
   {0xFB54, 0xFB54, 4167}, // {L'T', L'T', 4167},
   {0xFB55, 0xFB55, 4192}, // {L'U', L'U', 4192},
   {0xFB56, 0xFB56, 4184}, // {L'V', L'V', 4184}, /*?*/
   {0xFB57, 0xFB57, 4185}, // {L'W', L'W', 4185},
   {0xFB58, 0xE086, 0000}, // {L'X', L'', 0000},
   {0xFB59, 0xFB59, 4219}, // {L'Y', L'Y', 4219}, /*?*/
   {0xFB5A, 0xFB5A, 4229}, // {L'Z', L'Z', 4229}, /*?*/
   {0xFB5B, 0xFB5B, 4279}, // {L'[', L'[', 4279}, /*?*/
   {0xFB5C, 0xFB5C, 4281}, // {L'\', L'\', 4281}, /*?*/
   {0xFB5D, 0xFB5D, 4314}, // {L']', L']', 4314}, /*?*/
   {0xFB5E, 0x9296, 0000}, // {L'^', L'', 0000},
   {0xFB5F, 0xFB5F, 4342}, // {L'_', L'_', 4342},
   {0xFB60, 0xFB60, 4370}, // {L'`', L'`', 4370}, /*?*/
   {0xFB61, 0xFB61, 4374}, // {L'a', L'a', 4374},
   {0xFB62, 0xFB62, 4380}, // {L'b', L'b', 4380}, /*?*/
   {0xFB63, 0xFB63, 4384}, // {L'c', L'c', 4384}, /*?*/
   {0xFB64, 0xFB64, 4378}, // {L'd', L'd', 4378}, /*?*/
   {0xFB65, 0xFB65, 4405}, // {L'e', L'e', 4405}, /*?*/
   {0xFB66, 0xFB66, 4389}, // {L'f', L'f', 4389}, /*?*/
   {0xFB67, 0xFB67, 4411}, // {L'g', L'g', 4411},
   {0xFB68, 0xFB68, 4414}, // {L'h', L'h', 4414},
   {0xFB69, 0xFB69, 4413}, // {L'i', L'i', 4413}, /*?*/
   {0xFB6A, 0xFB6A, 4418}, // {L'j', L'j', 4418},
   {0xFB6B, 0xFB6B, 4434}, // {L'k', L'k', 4434}, /*?*/
   {0xFB6C, 0xFB6C, 4448}, // {L'l', L'l', 4448}, /*?*/
   {0xFB6D, 0xFB6D, 4457}, // {L'm', L'm', 4457}, /*?*/
   {0xFB6E, 0x9572, 0000}, // {L'n', L'r', 0000},
   {0xFB6F, 0xFB6F, 4527}, // {L'o', L'o', 4527},
   {0xFB70, 0xE1A3, 0000}, // {L'p', L'', 0000},
   {0xFB71, 0xFB71, 4627}, // {L'q', L'q', 4627},
   {0xFB72, 0xFB72, 0000}, // {L'r', L'r', 0000},
   {0xFB73, 0xFB73, 4626}, // {L's', L's', 4626}, /*?*/
   {0xFB74, 0xFB74, 4635}, // {L't', L't', 4635},
   {0xFB75, 0x8976, 0000}, // {L'u', L'v', 0000},
   {0xFB76, 0xFB76, 4682}, // {L'v', L'v', 4682},
   {0xFB77, 0xFB77, 1961}, // {L'w', L'w', 1961}, /*?*/
   {0xFB78, 0xFB78, 4735}, // {L'x', L'x', 4735}, /*?*/
   {0xFB79, 0xFB79, 4766}, // {L'y', L'y', 4766}, /*?*/
   {0xFB7A, 0xFB7A, 4773}, // {L'z', L'z', 4773}, /*?*/
   {0xFB7B, 0xFB7B, 4777}, // {L'{', L'{', 4777}, /*?*/
   {0xFB7C, 0xFB7C, 0000}, // {L'|', L'|', 0000}, /*?*/
   {0xFB7D, 0x97E7, 0000}, // {L'}', L'', 0000},
   {0xFB7E, 0x905F, 0000}, // {L'~', L'_', 0000},
   {0xFB80, 0x8FCB, 0000}, // {L'', L'', 0000},
   {0xFB81, 0xFB81, 4869}, // {L'', L'', 4869}, /*?*/
   {0xFB82, 0x959F, 0000}, // {L'', L'', 0000},
   {0xFB83, 0xFB83, 4873}, // {L'', L'', 4873}, /*?*/
   {0xFB84, 0xFB84, 4972}, // {L'', L'', 4972},
   {0xFB85, 0xFB85, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFB86, 0x96F5, 0000}, // {L'', L'', 0000},
   {0xFB87, 0x96F5, 4976}, // {L'', L'', 4976},
   {0xFB88, 0xFB88, 5027}, // {L'', L'', 5027}, /*?*/
   {0xFB89, 0x90B8, 0000}, // {L'', L'', 0000},
   {0xFB8A, 0xFB8A, 5165}, // {L'', L'', 5165}, /*?*/
   {0xFB8B, 0xFB8B, 5172}, // {L'', L'', 5172},
   {0xFB8C, 0xFB8C, 5207}, // {L'', L'', 5207}, /*?*/
   {0xFB8D, 0x97CE, 0000}, // {L'', L'', 0000},
   {0xFB8E, 0x8F8F, 0000}, // {L'', L'', 0000},
   {0xFB8F, 0xFB8F, 5252}, // {L'', L'', 5252},
   {0xFB90, 0x924D, 5282}, // {L'', L'M', 5282},
   {0xFB91, 0xFB91, 5314}, // {L'', L'', 5314}, /*?*/
   {0xFB92, 0x8948, 0000}, // {L'', L'H', 0000},
   {0xFB93, 0xFB93, 5562}, // {L'', L'', 5562},
   {0xFB94, 0xFB94, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFB95, 0xFB95, 5602}, // {L'', L'', 5602}, /*?*/
   {0xFB96, 0x8CD4, 5623}, // {L'', L'', 5623},
   {0xFB97, 0xFB97, 5633}, // {L'', L'', 5633}, /*?*/
   {0xFB98, 0xFB98, 5649}, // {L'', L'', 5649}, /*?*/
   {0xFB99, 0xFB99, 5684}, // {L'', L'', 5684}, /*?*/
   {0xFB9A, 0xFB9A, 5738}, // {L'', L'', 5738}, /*?*/
   {0xFB9B, 0xFB9B, 5740}, // {L'', L'', 5740},
   {0xFB9C, 0xFB9C, 5749}, // {L'', L'', 5749}, /*?*/
   {0xFB9D, 0xE463, 0000}, // {L'', L'c', 0000},
   {0xFB9E, 0x8C4F, 0000}, // {L'', L'O', 0000},
   {0xFB9F, 0xFB9F, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFBA0, 0xFBA0, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFBA1, 0x8A61, 0000}, // {L'', L'a', 0000},
   {0xFBA2, 0xE5E8, 6051}, // {L'', L'', 6051},
   {0xFBA3, 0xFBA3, 6140}, // {L'', L'', 6140},
   {0xFBA4, 0xFBA4, 6152}, // {L'', L'', 6152}, /*?*/
   {0xFBA5, 0xFBA5, 6174}, // {L'', L'', 6174},
   {0xFBA6, 0xFBA6, 6190}, // {L'', L'', 6190}, /*?*/
   {0xFBA7, 0xFBA7, 6204}, // {L'', L'', 6204},
   {0xFBA8, 0xFBA8, 6216}, // {L'', L'', 6216}, /*?*/
   {0xFBA9, 0x8F94, 0000}, // {L'', L'', 0000},
   {0xFBAA, 0xFBAA, 6221}, // {L'', L'', 6221}, /*?*/
   {0xFBAB, 0xFBAB, 6243}, // {L'', L'', 6243}, /*?*/
   {0xFBAC, 0xFBAC, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFBAD, 0xFBAD, 6325}, // {L'', L'', 6325}, /*?*/
   {0xFBAE, 0x978A, 0000}, // {L'', L'', 0000},
   {0xFBAF, 0x8CAB, 6335}, // {L'', L'', 6335},
   {0xFBB0, 0xFBB0, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFBB1, 0xFBB1, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFBB2, 0xFBB2, 6457}, // {L'', L'', 6457},
   {0xFBB3, 0xFBB3, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFBB4, 0x88ED, 0000}, // {L'', L'', 0000},
   {0xFBB5, 0xFBB5, 6566}, // {L'', L'', 6566}, /*?*/
   {0xFBB6, 0x9859, 0000}, // {L'', L'Y', 0000},
   {0xFBB7, 0x9373, 0000}, // {L'', L's', 0000},
   {0xFBB8, 0x8BBD, 0000}, // {L'', L'', 0000},
   {0xFBB9, 0xFBB9, 6639}, // {L'', L'', 6639},
   {0xFBBA, 0xFBBA, 6708}, // {L'', L'', 6708}, /*?*/
   {0xFBBB, 0xFBBB, 6706}, // {L'', L'', 6706}, /*?*/
   {0xFBBC, 0xFBBC, 6709}, // {L'', L'', 6709}, /*?*/
   {0xFBBD, 0xFBBD, 6715}, // {L'', L'', 6715}, /*?*/
   {0xFBBE, 0xFBBE, 6716}, // {L'', L'', 6716}, /*?*/
   {0xFBBF, 0xFBBF, 6710}, // {L'', L'', 6710}, /*?*/
   {0xFBC0, 0xFBC0, 6711}, // {L'', L'', 6711}, /*?*/
   {0xFBC1, 0x8994, 6728}, // {L'', L'', 6728},
   {0xFBC2, 0xFBC2, 6733}, // {L'', L'', 6733},
   {0xFBC3, 0xFBC3, 6731}, // {L'', L'', 6731}, /*?*/
   {0xFBC4, 0xFBC4, 6753}, // {L'', L'', 6753}, /*?*/
   {0xFBC5, 0xFBC5, 6756}, // {L'', L'', 6756}, /*?*/
   {0xFBC6, 0xFBC6, 6754}, // {L'', L'', 6754}, /*?*/
   {0xFBC7, 0xFBC7, 6763}, // {L'', L'', 6763}, /*?*/
   {0xFBC8, 0xFBC8, 6767}, // {L'', L'', 6767}, /*?*/
   {0xFBC9, 0xFBC9, 6765}, // {L'', L'', 6765}, /*?*/
   {0xFBCA, 0xFBCA, 6752}, // {L'', L'', 6752}, /*?*/
   {0xFBCB, 0xFBCB, 6773}, // {L'', L'', 6773}, /*?*/
   {0xFBCC, 0xFBCC, 6806}, // {L'', L'', 6806}, /*?*/
   {0xFBCD, 0xFBCD, 6781}, // {L'', L'', 6781}, /*?*/
   {0xFBCE, 0xFBCE, 6782}, // {L'', L'', 6782}, /*?*/
   {0xFBCF, 0xFBCF, 6843}, // {L'', L'', 6843}, /*?*/
   {0xFBD0, 0xFBD0, 6833}, // {L'', L'', 6833},
   {0xFBD1, 0xFBD1, 6835}, // {L'', L'', 6835}, /*?*/
   {0xFBD2, 0xFBD2, 6830}, // {L'', L'', 6830}, /*?*/
   {0xFBD3, 0xFBD3, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFBD4, 0xFBD4, 6832}, // {L'', L'', 6832}, /*?*/
   {0xFBD5, 0xFBD5, 6839}, // {L'', L'', 6839}, /*?*/
   {0xFBD6, 0xFBD6, 6831}, // {L'', L'', 6831}, /*?*/
   {0xFBD7, 0xFBD7, 6864}, // {L'', L'', 6864}, /*?*/
   {0xFBD8, 0xFBD8, 6862}, // {L'', L'', 6862},
   {0xFBD9, 0xFBD9, 6849}, // {L'', L'', 6849}, /*?*/
   {0xFBDA, 0xFBDA, 6860}, // {L'', L'', 6860}, /*?*/
   {0xFBDB, 0xFBDB, 0000}, // {L'', L'', 0000},
   {0xFBDC, 0xFBDC, 6850}, // {L'', L'', 6850}, /*?*/
   {0xFBDD, 0xFBDD, 6859}, // {L'', L'', 6859}, /*?*/
   {0xFBDE, 0xFBDE, 6852}, // {L'', L'', 6852}, /*?*/
   {0xFBDF, 0xFBDF, 6890}, // {L'', L'', 6890}, /*?*/
   {0xFBE0, 0xFBE0, 6879}, // {L'', L'', 6879}, /*?*/
   {0xFBE1, 0xFBE1, 6919}, // {L'', L'', 6919}, /*?*/
   {0xFBE2, 0xFBE2, 6931}, // {L'', L'', 6931}, /*?*/
   {0xFBE3, 0xFBE3, 6941}, // {L'', L'', 6941}, /*?*/
   {0xFBE4, 0xFBE4, 6948}, // {L'', L'', 6948}, /*?*/
   {0xFBE5, 0xFBE5, 6966}, // {L'', L'', 6966}, /*?*/
   {0xFBE6, 0xFBE6, 6974}, // {L'', L'', 6974}, /*?*/
   {0xFBE7, 0xFBE7, 6975}, // {L'', L'', 6975}, /*?*/
   {0xFBE8, 0x8AD4, 0000}, // {L'', L'', 0000},
   {0xFBE9, 0x97B2, 0000}, // {L'', L'', 0000},
   {0xFBEA, 0xFBEA, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFBEB, 0xFBEB, 7072}, // {L'', L'', 7072}, /*?*/
   {0xFBEC, 0x9388, 7079}, // {L'', L'', 7079},
   {0xFBED, 0xFBED, 7112}, // {L'', L'', 7112}, /*?*/
   {0xFBEE, 0xFBEE, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFBEF, 0xFBEF, 7114}, // {L'', L'', 7114}, /*?*/
   {0xFBF0, 0xFBF0, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFBF1, 0xFBF1, 7117}, // {L'', L'', 7117}, /*?*/
   {0xFBF2, 0x90C2, 0000}, // {L'', L'', 0000},
   {0xFBF3, 0xFBF3, 7118}, // {L'', L'', 7118}, /*?*/
   {0xFBF4, 0xFBF4, 7217}, // {L'', L'', 7217}, /*?*/
   {0xFBF5, 0xFBF5, 7222}, // {L'', L'', 7222}, /*?*/
   {0xFBF6, 0x94D1, 0000}, // {L'', L'', 0000},
   {0xFBF7, 0x8E94, 0000}, // {L'', L'', 0000},
   {0xFBF8, 0xFBF8, 7261}, // {L'', L'', 7261}, /*?*/
   {0xFBF9, 0x8AD9, 0000}, // {L'', L'', 0000},
   {0xFBFA, 0xFBFA, 0000}, // {L'', L'', 0000}, /*?*/
   {0xFBFB, 0xFBFB, 7349}, // {L'', L'', 7349}, /*?*/
   {0xFBFC, 0x8D82, 0000}, // {L'', L'', 0000},
   {0xFC40, 0xFC40, 7377}, // {L'@', L'@', 7377}, /*?*/
   {0xFC41, 0xFC41, 7445}, // {L'A', L'A', 7445}, /*?*/
   {0xFC42, 0xFC42, 0000}, // {L'B', L'B', 0000}, /*?*/
   {0xFC43, 0xFC43, 7459}, // {L'C', L'C', 7459}, /*?*/
   {0xFC44, 0xFC44, 7470}, // {L'D', L'D', 7470},
   {0xFC45, 0xFC45, 7474}, // {L'E', L'E', 7474}, /*?*/
   {0xFC46, 0xFC46, 7505}, // {L'F', L'F', 7505}, /*?*/
   {0xFC47, 0xE8B8, 7591}, // {L'G', L'', 7591},
   {0xFC48, 0xFC48, 7590}, // {L'H', L'H', 7590}, /*?*/
   {0xFC49, 0x92DF, 0000}, // {L'I', L'', 0000},
   {0xFC4A, 0xFC4A, 7654}, // {L'J', L'J', 7654}, /*?*/
   {0xFC4B, 0x8D95, 0000}, // {L'K', L'', 0000}
};

struct DbcsConvTable JisDbcsConvTable1[] = {
   {0x8345, 0x8394}, // {L'E', L''},
   {0x82A9, 0x82AA}, // {L'', L''},
   {0x82AB, 0x82AC}, // {L'', L''},
   {0x82AD, 0x82AE}, // {L'', L''},
   {0x82AF, 0x82B0}, // {L'', L''},
   {0x82B1, 0x82B2}, // {L'', L''},
   {0x834A, 0x834B}, // {L'J', L'K'},
   {0x834C, 0x834D}, // {L'L', L'M'},
   {0x834E, 0x834F}, // {L'N', L'O'},
   {0x8350, 0x8351}, // {L'P', L'Q'},
   {0x8352, 0x8353}, // {L'R', L'S'},
   {0x82B3, 0x82B4}, // {L'', L''},
   {0x82B5, 0x82B6}, // {L'', L''},
   {0x82B7, 0x82B8}, // {L'', L''},
   {0x82B9, 0x82BA}, // {L'', L''},
   {0x82BB, 0x82BC}, // {L'', L''},
   {0x8354, 0x8355}, // {L'T', L'U'},
   {0x8356, 0x8357}, // {L'V', L'W'},
   {0x8358, 0x8359}, // {L'X', L'Y'},
   {0x835A, 0x835B}, // {L'Z', L'['},
   {0x835C, 0x835D}, // {L'\', L']'},
   {0x82BD, 0x82BE}, // {L'', L''},
   {0x82BF, 0x82C0}, // {L'', L''},
   {0x82C2, 0x82C3}, // {L'', L''},
   {0x82C4, 0x82C5}, // {L'', L''},
   {0x82C6, 0x82C7}, // {L'', L''},
   {0x835E, 0x835F}, // {L'^', L'_'},
   {0x8360, 0x8361}, // {L'`', L'a'},
   {0x8363, 0x8364}, // {L'c', L'd'},
   {0x8365, 0x8366}, // {L'e', L'f'},
   {0x8367, 0x8368}, // {L'g', L'h'},
   {0x82CD, 0x82CE}, // {L'', L''},
   {0x82D0, 0x82D1}, // {L'', L''},
   {0x82D3, 0x82D4}, // {L'', L''},
   {0x82D6, 0x82D7}, // {L'', L''},
   {0x82D9, 0x82DA}, // {L'', L''},
   {0x836E, 0x836F}, // {L'n', L'o'},
   {0x8371, 0x8372}, // {L'q', L'r'},
   {0x8374, 0x8375}, // {L't', L'u'},
   {0x8377, 0x8378}, // {L'w', L'x'},
   {0x837A, 0x837B}, // {L'z', L'{'}
};

struct DbcsConvTable JisDbcsConvTable2[] = {
   {0x82CD, 0x82CF}, // {L'', L''},
   {0x82D0, 0x82D2}, // {L'', L''},
   {0x82D3, 0x82D5}, // {L'', L''},
   {0x82D6, 0x82D8}, // {L'', L''},
   {0x82D9, 0x82DB}, // {L'', L''},
   {0x836E, 0x8370}, // {L'n', L'p'},
   {0x8371, 0x8373}, // {L'q', L's'},
   {0x8374, 0x8376}, // {L't', L'v'},
   {0x8377, 0x8379}, // {L'w', L'y'},
   {0x837A, 0x837C}, // {L'z', L'|'}
};

struct CaseTable {
   wchar_t db;
   wchar_t dbupper;
   PCHAR sb;
   PCHAR sbupper;
};

struct CaseTable JisCaseTable[] = {
   {0x8140, 0x8140, " ", " "}, // {L'@', L'@', " ", " "},
   {0x8149, 0x8149, "!", "!"}, // {L'I', L'I', "!", "!"},
   {0xFA57, 0xFA57, "\"", "\""}, // {L'W', L'W', "\"", "\""},
   {0x8167, 0x8167, "\"", "\""}, // {L'g', L'g', "\"", "\""},
   {0x8168, 0x8168, "\"", "\""}, // {L'h', L'h', "\"", "\""},
   {0x8194, 0x8194, "#", "#"}, // {L'', L'', "#", "#"},
   {0x8190, 0x8190, "$", "$"}, // {L'', L'', "$", "$"},
   {0x8193, 0x8193, "%", "%"}, // {L'', L'', "%", "%"},
   {0x8195, 0x8195, "&", "&"}, // {L'', L'', "&", "&"},
   {0xFA56, 0xFA56, "'", "'"}, // {L'V', L'V', "'", "'"},
   {0x8165, 0x8165, "'", "'"}, // {L'e', L'e', "'", "'"},
   {0x8166, 0x8166, "'", "'"}, // {L'f', L'f', "'", "'"},
   {0x8169, 0x8169, "(", "("}, // {L'i', L'i', "(", "("},
   {0x816A, 0x816A, ")", ")"}, // {L'j', L'j', ")", ")"},
   {0x8196, 0x8196, "*", "*"}, // {L'', L'', "*", "*"},
   {0x817B, 0x817B, "+", "+"}, // {L'{', L'{', "+", "+"},
   {0x8143, 0x8143, ",", ","}, // {L'C', L'C', ",", ","},
   {0x817C, 0x817C, "-", "-"}, // {L'|', L'|', "-", "-"},
   {0x8144, 0x8144, ".", "."}, // {L'D', L'D', ".", "."},
   {0x815E, 0x815E, "/", "/"}, // {L'^', L'^', "/", "/"},
   {0x824F, 0x824F, "0", "0"}, // {L'O', L'O', "0", "0"},
   {0x8250, 0x8250, "1", "1"}, // {L'P', L'P', "1", "1"},
   {0x8251, 0x8251, "2", "2"}, // {L'Q', L'Q', "2", "2"},
   {0x8252, 0x8252, "3", "3"}, // {L'R', L'R', "3", "3"},
   {0x8253, 0x8253, "4", "4"}, // {L'S', L'S', "4", "4"},
   {0x8254, 0x8254, "5", "5"}, // {L'T', L'T', "5", "5"},
   {0x8255, 0x8255, "6", "6"}, // {L'U', L'U', "6", "6"},
   {0x8256, 0x8256, "7", "7"}, // {L'V', L'V', "7", "7"},
   {0x8257, 0x8257, "8", "8"}, // {L'W', L'W', "8", "8"},
   {0x8258, 0x8258, "9", "9"}, // {L'X', L'X', "9", "9"},
   {0x8146, 0x8146, ":", ":"}, // {L'F', L'F', ":", ":"},
   {0x8147, 0x8147, ";", ";"}, // {L'G', L'G', ";", ";"},
   {0x8183, 0x8183, "<", "<"}, // {L'', L'', "<", "<"},
   {0x8181, 0x8181, "=", "="}, // {L'', L'', "=", "="},
   {0x8184, 0x8184, ">", ">"}, // {L'', L'', ">", ">"},
   {0x8148, 0x8148, "?", "?"}, // {L'H', L'H', "?", "?"},
   {0x8197, 0x8197, "@", "@"}, // {L'', L'', "@", "@"},
   {0x8260, 0x8260, "A", "A"}, // {L'`', L'`', "A", "A"},
   {0x8261, 0x8261, "B", "B"}, // {L'a', L'a', "B", "B"},
   {0x8262, 0x8262, "C", "C"}, // {L'b', L'b', "C", "C"},
   {0x8263, 0x8263, "D", "D"}, // {L'c', L'c', "D", "D"},
   {0x8264, 0x8264, "E", "E"}, // {L'd', L'd', "E", "E"},
   {0x8265, 0x8265, "F", "F"}, // {L'e', L'e', "F", "F"},
   {0x8266, 0x8266, "G", "G"}, // {L'f', L'f', "G", "G"},
   {0x8267, 0x8267, "H", "H"}, // {L'g', L'g', "H", "H"},
   {0x8268, 0x8268, "I", "I"}, // {L'h', L'h', "I", "I"},
   {0x8269, 0x8269, "J", "J"}, // {L'i', L'i', "J", "J"},
   {0x826A, 0x826A, "K", "K"}, // {L'j', L'j', "K", "K"},
   {0x826B, 0x826B, "L", "L"}, // {L'k', L'k', "L", "L"},
   {0x826C, 0x826C, "M", "M"}, // {L'l', L'l', "M", "M"},
   {0x826D, 0x826D, "N", "N"}, // {L'm', L'm', "N", "N"},
   {0x826E, 0x826E, "O", "O"}, // {L'n', L'n', "O", "O"},
   {0x826F, 0x826F, "P", "P"}, // {L'o', L'o', "P", "P"},
   {0x8270, 0x8270, "Q", "Q"}, // {L'p', L'p', "Q", "Q"},
   {0x8271, 0x8271, "R", "R"}, // {L'q', L'q', "R", "R"},
   {0x8272, 0x8272, "S", "S"}, // {L'r', L'r', "S", "S"},
   {0x8273, 0x8273, "T", "T"}, // {L's', L's', "T", "T"},
   {0x8274, 0x8274, "U", "U"}, // {L't', L't', "U", "U"},
   {0x8275, 0x8275, "V", "V"}, // {L'u', L'u', "V", "V"},
   {0x8276, 0x8276, "W", "W"}, // {L'v', L'v', "W", "W"},
   {0x8277, 0x8277, "X", "X"}, // {L'w', L'w', "X", "X"},
   {0x8278, 0x8278, "Y", "Y"}, // {L'x', L'x', "Y", "Y"},
   {0x8279, 0x8279, "Z", "Z"}, // {L'y', L'y', "Z", "Z"},
   {0x816D, 0x816D, "[", "["}, // {L'm', L'm', "[", "["},
   {0x818F, 0x818F, "\\", "\\"}, // {L'', L'', "\\", "\\"},
   {0x816E, 0x816E, "]", "]"}, // {L'n', L'n', "]", "]"},
   {0x814F, 0x814F, "^", "^"}, // {L'O', L'O', "^", "^"},
   {0x8151, 0x8151, "_", "_"}, // {L'Q', L'Q', "_", "_"},
   {0x814D, 0x814D, "`", "`"}, // {L'M', L'M', "`", "`"},
   {0x8281, 0x8260, "a", "A"}, // {L'', L'`', "a", "A"},
   {0x8282, 0x8261, "b", "B"}, // {L'', L'a', "b", "B"},
   {0x8283, 0x8262, "c", "C"}, // {L'', L'b', "c", "C"},
   {0x8284, 0x8263, "d", "D"}, // {L'', L'c', "d", "D"},
   {0x8285, 0x8264, "e", "E"}, // {L'', L'd', "e", "E"},
   {0x8286, 0x8265, "f", "F"}, // {L'', L'e', "f", "F"},
   {0x8287, 0x8266, "g", "G"}, // {L'', L'f', "g", "G"},
   {0x8288, 0x8267, "h", "H"}, // {L'', L'g', "h", "H"},
   {0x8289, 0x8268, "i", "I"}, // {L'', L'h', "i", "I"},
   {0x828A, 0x8269, "j", "J"}, // {L'', L'i', "j", "J"},
   {0x828B, 0x826A, "k", "K"}, // {L'', L'j', "k", "K"},
   {0x828C, 0x826B, "l", "L"}, // {L'', L'k', "l", "L"},
   {0x828D, 0x826C, "m", "M"}, // {L'', L'l', "m", "M"},
   {0x828E, 0x826D, "n", "N"}, // {L'', L'm', "n", "N"},
   {0x828F, 0x826E, "o", "O"}, // {L'', L'n', "o", "O"},
   {0x8290, 0x826F, "p", "P"}, // {L'', L'o', "p", "P"},
   {0x8291, 0x8270, "q", "Q"}, // {L'', L'p', "q", "Q"},
   {0x8292, 0x8271, "r", "R"}, // {L'', L'q', "r", "R"},
   {0x8293, 0x8272, "s", "S"}, // {L'', L'r', "s", "S"},
   {0x8294, 0x8273, "t", "T"}, // {L'', L's', "t", "T"},
   {0x8295, 0x8274, "u", "U"}, // {L'', L't', "u", "U"},
   {0x8296, 0x8275, "v", "V"}, // {L'', L'u', "v", "V"},
   {0x8297, 0x8276, "w", "W"}, // {L'', L'v', "w", "W"},
   {0x8298, 0x8277, "x", "X"}, // {L'', L'w', "x", "X"},
   {0x8299, 0x8278, "y", "Y"}, // {L'', L'x', "y", "Y"},
   {0x829A, 0x8279, "z", "Z"}, // {L'', L'y', "z", "Z"},
   {0x816F, 0x816F, "{", "{"}, // {L'o', L'o', "{", "{"},
   {0x8162, 0x8162, "|", "|"}, // {L'b', L'b', "|", "|"},
   {0x8170, 0x8170, "}", "}"}, // {L'p', L'p', "}", "}"},
   {0x8150, 0x8150, "~", "~"}, // {L'P', L'P', "~", "~"},
   {0x8142, 0x8142, "", ""}, // {L'B', L'B', "", ""},
   {0x8175, 0x8175, "", ""}, // {L'u', L'u', "", ""},
   {0x8176, 0x8176, "", ""}, // {L'v', L'v', "", ""},
   {0x8141, 0x8141, "", ""}, // {L'A', L'A', "", ""},
   {0x8145, 0x8145, "", ""}, // {L'E', L'E', "", ""},
   {0x8392, 0x8392, "", ""}, // {L'', L'', "", ""},
   {0x8340, 0x8341, "", ""}, // {L'@', L'A', "", ""},
   {0x8342, 0x8343, "", ""}, // {L'B', L'C', "", ""},
   {0x8344, 0x8345, "", ""}, // {L'D', L'E', "", ""},
   {0x8346, 0x8347, "", ""}, // {L'F', L'G', "", ""},
   {0x8348, 0x8349, "", ""}, // {L'H', L'I', "", ""},
   {0x8383, 0x8384, "", ""}, // {L'', L'', "", ""},
   {0x8385, 0x8386, "", ""}, // {L'', L'', "", ""},
   {0x8387, 0x8388, "", ""}, // {L'', L'', "", ""},
   {0x8362, 0x8363, "", ""}, // {L'b', L'c', "", ""},
   {0x815B, 0x815B, "", ""}, // {L'[', L'[', "", ""},
   {0x8341, 0x8341, "", ""}, // {L'A', L'A', "", ""},
   {0x8343, 0x8343, "", ""}, // {L'C', L'C', "", ""},
   {0x8345, 0x8345, "", ""}, // {L'E', L'E', "", ""},
   {0x8394, 0x8374, "", ""}, // {L'', L't', "", ""},
   {0x8347, 0x8347, "", ""}, // {L'G', L'G', "", ""},
   {0x8349, 0x8349, "", ""}, // {L'I', L'I', "", ""},
   {0x834A, 0x834A, "", ""}, // {L'J', L'J', "", ""},
   {0x834C, 0x834C, "", ""}, // {L'L', L'L', "", ""},
   {0x834E, 0x834E, "", ""}, // {L'N', L'N', "", ""},
   {0x8350, 0x8350, "", ""}, // {L'P', L'P', "", ""},
   {0x8352, 0x8352, "", ""}, // {L'R', L'R', "", ""},
   {0x834B, 0x834A, "", ""}, // {L'K', L'J', "", ""},
   {0x834D, 0x834C, "", ""}, // {L'M', L'L', "", ""},
   {0x834F, 0x834E, "", ""}, // {L'O', L'N', "", ""},
   {0x8351, 0x8350, "", ""}, // {L'Q', L'P', "", ""},
   {0x8353, 0x8352, "", ""}, // {L'S', L'R', "", ""},
   {0x8354, 0x8354, "", ""}, // {L'T', L'T', "", ""},
   {0x8356, 0x8356, "", ""}, // {L'V', L'V', "", ""},
   {0x8358, 0x8358, "", ""}, // {L'X', L'X', "", ""},
   {0x835A, 0x835A, "", ""}, // {L'Z', L'Z', "", ""},
   {0x835C, 0x835C, "", ""}, // {L'\', L'\', "", ""},
   {0x8355, 0x8354, "", ""}, // {L'U', L'T', "", ""},
   {0x8357, 0x8356, "", ""}, // {L'W', L'V', "", ""},
   {0x8359, 0x8358, "", ""}, // {L'Y', L'X', "", ""},
   {0x835B, 0x835A, "", ""}, // {L'[', L'Z', "", ""},
   {0x835D, 0x835C, "", ""}, // {L']', L'\', "", ""},
   {0x835E, 0x835E, "", ""}, // {L'^', L'^', "", ""},
   {0x8360, 0x8360, "", ""}, // {L'`', L'`', "", ""},
   {0x8363, 0x8363, "", ""}, // {L'c', L'c', "", ""},
   {0x8365, 0x8365, "", ""}, // {L'e', L'e', "", ""},
   {0x8367, 0x8367, "", ""}, // {L'g', L'g', "", ""},
   {0x835F, 0x835E, "", ""}, // {L'_', L'^', "", ""},
   {0x8361, 0x8360, "", ""}, // {L'a', L'`', "", ""},
   {0x8364, 0x8363, "", ""}, // {L'd', L'c', "", ""},
   {0x8366, 0x8365, "", ""}, // {L'f', L'e', "", ""},
   {0x8368, 0x8367, "", ""}, // {L'h', L'g', "", ""},
   {0x8369, 0x8369, "", ""}, // {L'i', L'i', "", ""},
   {0x836A, 0x836A, "", ""}, // {L'j', L'j', "", ""},
   {0x836B, 0x836B, "", ""}, // {L'k', L'k', "", ""},
   {0x836C, 0x836C, "", ""}, // {L'l', L'l', "", ""},
   {0x836D, 0x836D, "", ""}, // {L'm', L'm', "", ""},
   {0x836E, 0x836E, "", ""}, // {L'n', L'n', "", ""},
   {0x8371, 0x8371, "", ""}, // {L'q', L'q', "", ""},
   {0x8374, 0x8374, "", ""}, // {L't', L't', "", ""},
   {0x8377, 0x8377, "", ""}, // {L'w', L'w', "", ""},
   {0x837A, 0x837A, "", ""}, // {L'z', L'z', "", ""},
   {0x836F, 0x836E, "", ""}, // {L'o', L'n', "", ""},
   {0x8372, 0x8371, "", ""}, // {L'r', L'q', "", ""},
   {0x8375, 0x8374, "", ""}, // {L'u', L't', "", ""},
   {0x8378, 0x8377, "", ""}, // {L'x', L'w', "", ""},
   {0x837B, 0x837A, "", ""}, // {L'{', L'z', "", ""},
   {0x8370, 0x836E, "", ""}, // {L'p', L'n', "", ""},
   {0x8373, 0x8371, "", ""}, // {L's', L'q', "", ""},
   {0x8376, 0x8374, "", ""}, // {L'v', L't', "", ""},
   {0x8379, 0x8377, "", ""}, // {L'y', L'w', "", ""},
   {0x837C, 0x837A, "", ""}, // {L'|', L'z', "", ""},
   {0x837D, 0x837D, "", ""}, // {L'}', L'}', "", ""},
   {0x837E, 0x837E, "", ""}, // {L'~', L'~', "", ""},
   {0x8380, 0x8380, "", ""}, // {L'', L'', "", ""},
   {0x8381, 0x8381, "", ""}, // {L'', L'', "", ""},
   {0x8382, 0x8382, "", ""}, // {L'', L'', "", ""},
   {0x8384, 0x8384, "", ""}, // {L'', L'', "", ""},
   {0x8386, 0x8386, "", ""}, // {L'', L'', "", ""},
   {0x8388, 0x8388, "", ""}, // {L'', L'', "", ""},
   {0x8389, 0x8389, "", ""}, // {L'', L'', "", ""},
   {0x838A, 0x838A, "", ""}, // {L'', L'', "", ""},
   {0x838B, 0x838B, "", ""}, // {L'', L'', "", ""},
   {0x838C, 0x838C, "", ""}, // {L'', L'', "", ""},
   {0x838D, 0x838D, "", ""}, // {L'', L'', "", ""},
   {0x838F, 0x838F, "", ""}, // {L'', L'', "", ""},
   {0x8393, 0x8393, "", ""}, // {L'', L'', "", ""},
   {0x814A, 0x814A, "", ""}, // {L'J', L'J', "", ""},
   {0x814B, 0x814B, "", ""}, // {L'K', L'K', "", ""},
   {0x829F, 0x82A0, NULL, NULL}, // {L'', L'', NULL, NULL},
   {0x82A1, 0x82A2, NULL, NULL}, // {L'', L'', NULL, NULL},
   {0x82A3, 0x82A4, NULL, NULL}, // {L'', L'', NULL, NULL},
   {0x82A5, 0x82A6, NULL, NULL}, // {L'', L'', NULL, NULL},
   {0x82A7, 0x82A8, NULL, NULL}, // {L'', L'', NULL, NULL},
   {0x82E1, 0x82E2, NULL, NULL}, // {L'', L'', NULL, NULL},
   {0x82E3, 0x82E4, NULL, NULL}, // {L'', L'', NULL, NULL},
   {0x82E5, 0x82E6, NULL, NULL}, // {L'', L'', NULL, NULL},
   {0x82C1, 0x82C2, NULL, NULL}, // {L'', L'', NULL, NULL},
   {0x82EC, 0x82ED, NULL, NULL}, // {L'', L'', NULL, NULL},
   {0x82EE, 0x82A2, NULL, NULL}, // {L'', L'', NULL, NULL},
   {0x82EF, 0x82A6, NULL, NULL}, // {L'', L'', NULL, NULL},
   {0x82F0, 0x82A8, NULL, NULL}, // {L'', L'', NULL, NULL},
   {0x838E, 0x838F, NULL, NULL}, // {L'', L'', NULL, NULL},
   {0x8390, 0x8343, NULL, NULL}, // {L'', L'C', NULL, NULL},
   {0x8391, 0x8347, NULL, NULL}, // {L'', L'G', NULL, NULL},
   {0x8392, 0x8349, NULL, NULL}, // {L'', L'I', NULL, NULL},
   {0x8394, 0x8374, NULL, NULL}, // {L'', L't', NULL, NULL},
   {0x8395, 0x834A, NULL, NULL}, // {L'', L'J', NULL, NULL},
   {0x8396, 0x8350, NULL, NULL}, // {L'', L'P', NULL, NULL},
};

enum CharSet {
   USASCII,
   ISO8859_1,  
   ISO8859_2,  
   ISO8859_3,  
   ISO8859_4,  
   ISO8859_5,  
   ISO8859_6,  
   ISO8859_7,  
   ISO8859_8,  
   ISO8859_9,  
   ISO2022KR, 
   ISO2022JP,  
   JISX0212_1990,
   HankakuKatakana,
   SS2         
};

struct EscapeSeq {
   PSZ seq;
   int len;
   CharSet charset; 
   PWCHAR_T* jistopccodetable; 
};

static const EscapeSeq DoubleByteSeqTable[] = {
   {"\x1b$B", 3, ISO2022JP, Jis83ToPccodeTable},  
   {"\x1b$A", 3, ISO2022JP, Jis83ToPccodeTable},  
   {"\x1b$@", 3, ISO2022JP, Jis78ToPccodeTable},  
   {"\x1b$@\x1b$B", 6, ISO2022JP, Jis90ToPccodeTable}, 
   {"\x1b(J", 3, USASCII, NULL},   
   {"\x1b$(D", 4, JISX0212_1990, JisX0212_1990ToPccodeTable},   
   {"\x1b(B", 3, USASCII, NULL},   
   {"\x1b$(C", 4, ISO2022KR, NULL}, 
   {"\x1b$)C", 4, ISO2022KR, NULL}, 
   {"\x1bN", 2, SS2, NULL},   
   {"\x1b.A", 3, ISO8859_1, NULL},   
   {"\x1b.F", 3, ISO8859_7, NULL},   
   {"\x1b(I", 3, HankakuKatakana, NULL},   
   {NULL, 0}
};

struct JisEntry {
   int kutencode;
   wchar_t pccode;
};

struct JisTableEntry {
   JisEntry oldjis;
   JisEntry newjis;
};

static struct JisTableEntry JisTable[] = { 
   {{1619, 0xe9cb}, {1619, 0x88b1}}, /*  */
   {{8245, 0x88b1}, {8245, 0xe9cb}}, /*  */
   {{1809, 0xe9f2}, {1809, 0x89a7}}, /*  */
   {{8284, 0x89a7}, {8284, 0xe9f2}}, /*  */
   {{1934, 0xe579}, {1934, 0x8a61}}, /* y */
   {{7358, 0x8a61}, {7358, 0xe579}}, /* a */
   {{1941, 0x9d98}, {1941, 0x8a68}}, /*  */
   {{5788, 0x8a68}, {5788, 0x9d98}}, /* h */
   {{1986, 0xe27d}, {1986, 0x8a96}}, /* } */
   {{6862, 0x8a96}, {6762, 0xe27d}}, /*  */
   {{2035, 0x9ff3}, {2035, 0x8ac1}}, /*  */
   {{6285, 0x8ac1}, {9285, 0x9ff3}}, /*  */
   {{2050, 0xe67c}, {2050, 0x8ad0}}, /* | */
   {{7561, 0x8ad0}, {7561, 0xe67c}}, /*  */
   {{2359, 0xe8f2}, {2359, 0x8c7a}}, /*  */
   {{8084, 0x8c7a}, {8084, 0xe8f2}}, /* z */
   {{2560, 0xe1e6}, {2560, 0x8d7b}}, /*  */
   {{6672, 0x8d7b}, {6672, 0xe1e6}}, /* { */
   {{2841, 0xe541}, {2841, 0x8ec7}}, /* A */
   {{7302, 0x8ec7}, {7302, 0xe541}}, /*  */
   {{3157, 0xe8d5}, {3157, 0x9078}}, /*  */
   {{8055, 0x9078}, {7302, 0xe8d5}}, /* x */
   {{3308, 0xe6cb}, {3308, 0x9147}}, /*  */
   {{7645, 0x9147}, {7645, 0xe6cb}}, /* G */
   {{3659, 0x9ae2}, {3659, 0x92d9}}, /*  */
   {{5268, 0x92d9}, {5268, 0x9ae2}}, /*  */
   {{3755, 0xe1e8}, {3755, 0x9376}}, /*  */
   {{6674, 0x9376}, {6674, 0xe1e8}}, /* v */
   {{3778, 0x9e8d}, {3778, 0x938e}}, /*  */
   {{5977, 0x938e}, {5977, 0x9e8d}}, /*  */
   {{3783, 0x9fb7}, {3783, 0x9393}}, /*  */
   {{6225, 0x9393}, {6225, 0x9fb7}}, /*  */
   {{3886, 0xe78e}, {3886, 0x93f4}}, /*  */
   {{7778, 0x93f4}, {7778, 0xe78e}}, /*  */
   {{3972, 0xe5a2}, {3972, 0x9488}}, /*  */
   {{7404, 0x9488}, {7404, 0xe5a2}}, /*  */
   {{4116, 0x9e77}, {4116, 0x954f}}, /* w */
   {{5956, 0x954f}, {5956, 0x9e77}}, /* O */
   {{4389, 0x98d4}, {4389, 0x9699}}, /*  */
   {{4854, 0x9699}, {4854, 0x98d4}}, /*  */
   {{4489, 0xe54d}, {4489, 0x96f7}}, /* M */
   {{7314, 0x96f7}, {7314, 0xe54d}}, /*  */
   {{4722, 0xe2c4}, {4722, 0x9855}}, /*  */
   {{6838, 0x9855}, {6838, 0xe2c4}}, /* U */
   {{2238, 0xea9f}, {2238, 0x8bc4}}, /*  */
   {{8401, 0x8bc4}, {8401, 0xea9f}}, /*  */
   {{4374, 0xeaa0}, {4374, 0x968a}}, /*  */
   {{8402, 0x968a}, {8402, 0xeaa0}}, /*  */
   {{4558, 0xeaa1}, {4558, 0x9779}}, /*  */
   {{8403, 0x9779}, {8403, 0xeaa1}}, /* y */
   {{6486, 0xeaa2}, {6486, 0xe0f4}}, /*  */
   {{8404, 0xe0f4}, {8404, 0xeaa2}}, /*  */
   {{0272, 0xfa5b}, {0272, 0x81e6}}, /* [ */
   {{0244, 0xfa54}, {0244, 0x81ca}}, /* T */
   {{2523, 0x8d56}, {11650, 0xfad0}}, /* V */
   {{11650, 0xfad0}, {2523, 0x8d56}}, /*  */
   {{6370, 0xe086}, {8406, 0xeaa4}}, /*  */
   {{0, 0xe086}, {6370, 0xe086}}, /*  */
   {{0, 0xeaa3}, {8405, 0xeaa3}} /*  */
};

static const size_t NoOfJisEntries = sizeof(JisTable) / sizeof(JisTable[0]); 

static BOOL InJis90 = FALSE;

struct MimeCharSet {
   PCHAR name;
   int namesize;
   CharSet charset;
};

static struct MimeCharSet MimeCharSetTable[] = {
   {"US-ASCII", 8, USASCII},
   {"ISO-8859-1", 10, ISO8859_1}, 
   {"ISO-8859-2", 10, ISO8859_2}, 
   {"ISO-8859-3", 10, ISO8859_3}, 
   {"ISO-8859-4", 10, ISO8859_4}, 
   {"ISO-8859-5", 10, ISO8859_5}, 
   {"ISO-8859-6", 10, ISO8859_6}, 
   {"ISO-8859-7", 10, ISO8859_7}, 
   {"ISO-8859-8", 10, ISO8859_8}, 
   {"ISO-8859-9", 10, ISO8859_9}, 
   {"ISO-2022-JP", 11, ISO2022JP},
   {"ISO-2022-JP-2", 13, ISO2022JP},
   {"ISO-2022-KR", 11, ISO2022KR}
};

static const size_t NoOfMimeCharSet = sizeof(MimeCharSetTable) / sizeof(MimeCharSetTable[0]);

static struct MimeCharSet *MimeEncodedWordTable;
static size_t NoOfMimeEncodedWord;

static CHAR Base64Table[UCHAR_MAX+1] = {
   Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error,
   Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error,
   Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, 62, Base64Error, Base64Error, Base64Error, 63,
   52, 53, 54, 55, 56, 57, 58, 59, 60, 61, Base64Error, Base64Error, Base64Error, Base64Pad, Base64Error, Base64Error,
   Base64Error, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14,
   15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error,
   Base64Error, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
   41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error,
   Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error,
   Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error,
   Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error,
   Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error,
   Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error,
   Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error,
   Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error,
   Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error, Base64Error
};
static CHAR QuotedPrintableTable[UCHAR_MAX+1];
static PCHAR ToBase64Table = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static EscapeSeq ToQuotedPrintableTable[UCHAR_MAX+1];
static BOOL DbcsVec[UCHAR_MAX+1];

static int _Optlink
compcharset(
const void *key,
const void *element) {
   MimeCharSet *v1 = (MimeCharSet*)key;
   MimeCharSet *v2 = (MimeCharSet*)element;
   return (strnicmp(v1->name, v2->name, v2->namesize));
}

static int
find1stbyte(
PCHAR str,
int pos) {
   if (pos <= 0 || pos > strlen(str)) {
      return (0);
   } /* endif */
   if (strlen(str) == 1) {
      return (1);
   } /* endif */
   for (int i = pos - 1; i >= 2; i -= 1) {
      if (!IsDbcs1st(str[i]) && !IsDbcs1st(str[i-1])) {
         return (i + 1);
      } /* endif */
      if (IsDbcs1st(str[i]) && !IsDbcs1st(str[i-1])) {
         return (i + 1);
      } /* endif */
      if (IsDbcs1st(str[i]) && IsDbcs1st(str[i-1]) && !IsDbcs1st(str[i-2])) {
         return (i - 1);
      } /* endif */
   } /* endfor */
   if (!IsDbcs1st(str[0])) {
      return (2);
   } /* endif */
   return (1);
}

static int
nmbstowcs(
wchar_t *target,
size_t targetsize,
PCHAR source,
size_t sourcesize) {
   size_t i = 0;
   size_t count = 0;
   while (i < sourcesize) {
      RxJisCountCheck(count, targetsize);
      mbtowc(target + count, source + i, sourcesize - i);
      if (target[count] > UCHAR_MAX) {
         i += 2;
      } else {
         i += 1;
      } /* endif */
      count += 1;
   } /* endwhile */
   RxJisCountCheck(count, targetsize);
   target[count] = '\0';
   return (count);
}

static wchar_t
PcToJis(
wchar_t c,
PWCHAR_T* t = PccodeToJis83Table) {
   int l = c & ByteMask;
   int h = c >> ByteBitWidth;
   if (!t[h]) {
      return (JisSpace);
   } /* endif */
   return (t[h][l]);
}

static wchar_t
JisToPc(
wchar_t c,
PWCHAR_T* t) {
   if (!t) {
      return (c);
   } /* endif */
   int l = c & ByteMask;
   int h = c >> ByteBitWidth;
   if (!t[h]) {
      return (L'@');
   } /* endif */
   return (t[h][l]);
}

static int
Base64Size(
PCHAR source,
int sourcesize = 0,
int linesize = 0) {
   if (sourcesize <= 0) { 
      sourcesize = strlen(source);
   } /* endif */
   int n = ByteBitWidth * sourcesize;
   n /= Base64BitWidth;
   if (n % Base64BitWidth) {
      n += 1;
   } /* endif */
   if (linesize > 0) {
      int l = n / linesize;
      if (n % linesize) {
         l += 1;
      } /* endif */
      n += 2 * l;
   } /* endif */
   n += MaxMimeLineSize;
   return (n);
}

static int
DbcsSize(
PCHAR source,
int sourcesize = 0,
int linesize = 0) {
   if (sourcesize <= 0) { 
      sourcesize = strlen(source);
   } /* endif */
   return (2 * sourcesize);
}

static int
QuotedPrintableSize(
PCHAR source,
int sourcesize = 0,
int linesize = 0) {
   if (sourcesize <= 0) { 
      sourcesize = strlen(source);
   } /* endif */
   int n = 3 * sourcesize;
   if (linesize > 0) {
      int l = n / linesize;
      if (n % linesize) {
         l += 1;
      } /* endif */
      n += 2 * l;
   } /* endif */
   n += linesize;
   return (n);
}

static int
JisSize(
PCHAR source,
int sourcesize = 0,
int linesize = 0) {
   if (sourcesize <= 0) { 
      sourcesize = strlen(source);
   } /* endif */
   int count = 0;
   int sourcecount = 0;
   BOOL indoublebyte = FALSE;
   while (sourcecount <= sourcesize) { 
      if (!IsDbcs1st(source[sourcecount])) {  
         if (indoublebyte) {
            indoublebyte = FALSE;
            count += Jis83OutSeq.len; 
         } /* endif */
         count += 1;
         sourcecount += 1;
      } else {
         if (!indoublebyte) {
            indoublebyte = TRUE;
            count += Jis83InSeq.len; 
         } /* endif */
         count += 2;
         sourcecount += 2;
      } /* endif */
   } /* endwhile */
   if (indoublebyte) {
      count += Jis83OutSeq.len; 
   } /* endif */
   return (count + 1);
}

static int
EucSize(
PCHAR source,
int sourcesize = 0,
int linesize = 0) {
   if (sourcesize <= 0) { 
      sourcesize = strlen(source);
   } /* endif */
   int count = 0;
   size_t sourcecount = 0;
   while (sourcecount <= sourcesize) {
      if (!IsDbcs1st(source[sourcecount])) {
         count += 1;
         sourcecount += 1;
      } else {
         count += 2;
         sourcecount += 2;
      } /* endif */
   } /* endwhile */
   return (count + 1);
}

int _Export 
JisWords(
PCHAR source,
int sourcesize) {
   if (0 == sourcesize) {
      sourcesize = strlen(source);
   } /* endif */
   ULONG pos = 0;
   int count = 0;
   for (; ; ) {
      ULONG startword;
      ULONG endword;
      ULONG nextword;
      BOOL rc = NlsFindWord(DefaultCodepage, source, sourcesize, pos, &startword, &endword, &nextword);
      if (!rc) {
         return (RxJisError);
      } /* endif */
      count += 1;
      if (nextword >= sourcesize) {
         break;
      } /* endif */
      pos = nextword;
   } /* endfor */
   return (count);
}

int _Export 
JisWord(
char* target, 
char* source, 
int targetsize, 
int position, 
int sourcesize) {
   if (!target || !source || targetsize <= 0 || position <= 0) {
      return (RxJisError);
   } /* endif */
   if (0 == sourcesize) {
      sourcesize = strlen(source);
   } /* endif */
   if (sourcesize < 0) {
      return (RxJisError);
   } /* endif */
   if (0 == sourcesize) {
      return (0);
   } /* endif */
   int count = 0;
   ULONG pos = 0;
   ULONG startword;
   ULONG endword;
   for (; ; ) {
      ULONG nextword;
      BOOL rc = NlsFindWord(DefaultCodepage, source, sourcesize, pos, &startword, &endword, &nextword);
      if (!rc) {
         return (RxJisError);
      } /* endif */
      if (++count == position) {
         break;
      } /* endif */
      if (nextword >= sourcesize) {
         return (0);
      } /* endif */
      pos = nextword;
   } /* endfor */
   if (endword < startword) {
      return (RxJisError);
   } /* endif */
   count = endword - startword;
   if (count >= targetsize) {
      return (RxJisError);
   } /* endif */
   memcpy(target, source + startword, count);
   target[count] = '\0';
   return (count);
}

int _Export 
JisAppend(
char* target, 
char* source, 
int targetsize, 
int targetlength,
int sourcesize) {
   if (!target || !source || targetsize <= 0) {
      return (RxJisError);
   } /* endif */
   if (sourcesize < 0) {
      return (RxJisError);
   } /* endif */
   if (0 == sourcesize) {
      sourcesize = strlen(source);
   } /* endif */
   if (0 == targetlength) {
      targetlength = strlen(target);
   } /* endif */
   if (targetlength + 1 + sourcesize + 1 > targetsize) {
      return (RxJisError);
   } /* endif */
   if (0 == sourcesize) {
      target[targetlength] = '\0';
      return (targetlength);
   } /* endif */
   if (0 == targetlength) {
      memcpy(target, source, sourcesize);
      target[sourcesize] = '\0';
      return (sourcesize);
   } /* endif */
   int i = find1stbyte(target, targetlength);
   for (; ; ) {
      if (i == targetlength || IsDbcs1st(target[i-1]) && i + 1 == targetlength) {
         break;
      } /* endif */
      if (IsDbcs1st(target[i-1])) {
         i += 2;
      } else {
         i += 1;
      } /* endif */
   } /* endfor */
   if (i == targetlength) {
      if (!IsDbcs1st(source[0]) && !isspace(source[0]) && i > 0 && !isspace(target[i-1])) {
         target[targetlength] = ' ';
         targetlength += 1;
      } /* endif */
   } /* endif */           
   memcpy(&target[targetlength], source, sourcesize);
   targetlength += sourcesize;
   target[targetlength] = '\0';
   return (targetlength);
}

int _Export 
JisWordIndex(
char* target, 
char* source, 
int targetsize, 
int position, 
int sourcesize) {
   if (!target || !source || targetsize <= 0 || position <= 0) {
      return (RxJisError);
   } /* endif */
   if (0 == sourcesize) {
      sourcesize = strlen(source);
   } /* endif */
   if (sourcesize < 0) {
      return (RxJisError);
   } /* endif */
   if (0 == sourcesize) {
      return (0);
   } /* endif */
   int count = 0;
   ULONG pos = 0;
   ULONG startword;
   ULONG endword;
   for (; ; ) {
      ULONG nextword;
      BOOL rc = NlsFindWord(DefaultCodepage, source, sourcesize, pos, &startword, &endword, &nextword);
      if (!rc) {
         return (RxJisError);
      } /* endif */
      if (++count == position) {
         break;
      } /* endif */
      if (nextword >= sourcesize) {
         return (0);
      } /* endif */
      pos = nextword;
   } /* endfor */
   return (pos + 1);
}

static int
ToEuc(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize = 0,
int linesize = 0) {
   RxJisArgCheck(source, target, targetsize, sourcesize);
   int count = 0;
   size_t sourcecount = 0;
   while (sourcecount < sourcesize) { 
      if (source[sourcecount] <= SCHAR_MAX) {
         RxJisCountCheck(count, targetsize);
         target[count++] = source[sourcecount++];
      } else if (!IsDbcs1st(source[sourcecount])) {
         RxJisCountCheck(count + 1, targetsize);
         target[count++] = EucSs2;
         target[count++] = source[sourcecount++];
      } else {
         RxJisCountCheck(count + 1, targetsize);
         int h = source[sourcecount++];
         int l = source[sourcecount++];
         wchar_t c = h << ByteBitWidth | l;                    
         c = PcToJis(c);                                       
         target[count++] = EucMsb | (c >> ByteBitWidth);       
         target[count++] = EucMsb | (c & ByteMask);            
      } /* endif */
   } /* endwhile */
   RxJisCountCheck(count, targetsize);
   target[count] = '\0';
   return (count);                      
}

static int
MimeJisTo(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize = 0,
int linesize = 0) {
   RxJisArgCheck(source, target, targetsize, sourcesize);
   if (targetsize < sourcesize) {
      return (RxJisError);
   } /* endif */
   int count = 0;
   PCHAR temp1 = new CHAR[targetsize];
   PCHAR temp2 = new CHAR[targetsize];
   if (!temp1 || !temp2) {
      return (RxJisError);
   } /* endif */
   int sourcecount = 0;
   while (sourcecount < sourcesize) {
      PCHAR p;
      PCHAR s;
      MimeCharSet k;
      MimeCharSet *r;
      int n;
      int c;
      switch (source[sourcecount]) {
      case '=':
         k.name = source + sourcecount;
         r = (MimeCharSet*)bsearch(&k, MimeEncodedWordTable, NoOfMimeEncodedWord, sizeof(MimeEncodedWordTable[0]), compcharset);
         if (!r) {
            RxJisCountCheck(count, targetsize);
            temp1[count++] = source[sourcecount++];
            continue;
         } /* endif */
         sourcecount += r->namesize;
         s = source + sourcecount;
         p = strstr(source + sourcecount, "?=");
         if (!p) {
            n = sourcesize - sourcecount;
         } else {
            n = p - s;
         } /* endif */
         if ('B' == r->name[r->namesize-2]) {
            c = Base64To(temp2, s, targetsize, n);
         } else {
            c = QuotedPrintableTo(temp2, s, targetsize, n);
         } /* endif */
         if (c < 0) {
           return (RxJisError);
         } /* endif */
         sourcecount += n + 2;
         memcpy(temp1 + count, temp2, c);
         RxJisCountCheck(count + c, targetsize);
         count += c;
         break;
      default: 
        RxJisCountCheck(count, targetsize);
        temp1[count++] = source[sourcecount++];
        break;
      } /* endswitch */
   } /* endwhile */
   RxJisCountCheck(count, targetsize);
   temp1[count] = '\0';
   count = ToPc(target, temp1, targetsize, count);
   delete temp1;
   delete temp2;
   return (count);
}

static int
EucTo(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize = 0,
int linesize = 0) {
   RxJisArgCheck(source, target, targetsize, sourcesize);
   int count = 0;
   size_t sourcecount = 0;
   while (sourcecount < sourcesize) {
      if (source[sourcecount] <= SCHAR_MAX) {
         RxJisCountCheck(count, targetsize);
         target[count++] = source[sourcecount++];
      } else if (EucSs2 == source[sourcecount]) {
         RxJisCountCheck(count, targetsize);
         sourcecount += 1;
         target[count++] = source[sourcecount++];
      } else if (EucSs3 == source[sourcecount]) {
         sourcecount += 1;        
         RxJisCountCheck(count + 1, targetsize);
         int h = source[sourcecount++] & EucMask; 
         int l = source[sourcecount++] & EucMask; 
         wchar_t c = h << ByteBitWidth | l;        
         c = JisToPc(c, JisX0212_1990ToPccodeTable);                           
         target[count++] = c >> ByteBitWidth;      
         target[count++] = c & ByteMask;           
      } else {
         RxJisCountCheck(count + 1, targetsize);
         int h = source[sourcecount++] & EucMask; 
         int l = source[sourcecount++] & EucMask; 
         wchar_t c = h << ByteBitWidth | l;        
         c = JisToPc(c, Jis83ToPccodeTable);                           
         target[count++] = c >> ByteBitWidth;      
         target[count++] = c & ByteMask;           
      } /* endif */
   } /* endwhile */
   RxJisCountCheck(count, targetsize);
   target[count] = '\0';
   return (count);
}

static int
ToJisChar(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize = 0,
int linesize = 0) {
   RxJisArgCheck(source, target, targetsize, sourcesize);
   if (!sourcesize) {
      return (0);
   } /* endif */
   if (targetsize <= 3) {
      return (RxJisError);
   } /* endif */
   wchar_t *temp = new wchar_t[sourcesize+1];
   if (!temp) {
      return (RxJisError);
   } /* endif */
   nmbstowcs(temp, sourcesize, source, sourcesize);
   if (temp[0] <= UCHAR_MAX) {
      delete temp;
      return (0);
   } /* endif */
   wchar_t c = PcCodeToJisCode(temp[0]);
   target[0] = c >> ByteBitWidth;
   target[1] = c & ByteMask;
   target[2] = '\0';
   delete temp;
   return (2);
}

static int
JisCharTo(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize = 0,
int linesize = 0) {
   RxJisArgCheck(source, target, targetsize, sourcesize);
   if (sourcesize < 2) {
      return (0);
   } /* endif */
   if (targetsize <= 3) {
      return (RxJisError);
   } /* endif */
   wchar_t c = JisToPc((source[0] << ByteBitWidth) | source[1], Jis83ToPccodeTable);
   target[0] = c >> ByteBitWidth;
   target[1] = c & ByteMask;
   target[2] = '\0';
   return (2);
}

static int
ToJis(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize = 0,
int linesize = 0) {
   RxJisArgCheck(source, target, targetsize, sourcesize);
   int count = 0;
   size_t sourcecount = 0; 
   while (sourcecount < sourcesize) { 
      if (!IsDbcs1st(source[sourcecount])) { 
         RxJisCountCheck(count, targetsize);
         target[count++] = source[sourcecount++]; 
      } else {
         RxJisCountCheck(count + Jis83InSeq.len, targetsize);
         strncpy(target + count, Jis83InSeq.seq, Jis83InSeq.len); 
         count += Jis83InSeq.len; 
         while (sourcecount < sourcesize && IsDbcs1st(source[sourcecount])) { 
            RxJisCountCheck(count + 2, targetsize);
            int h = source[sourcecount++];
            int l = source[sourcecount++];
            wchar_t c = h << ByteBitWidth | l;                    
            c = PcToJis(c);                                       
            target[count++] = c >> ByteBitWidth; 
            target[count++] = c & ByteMask;      
         } /* endwhile */
         RxJisCountCheck(count + Jis83OutSeq.len, targetsize);
         strncpy(target + count, Jis83OutSeq.seq, Jis83OutSeq.len); 
         count += Jis83OutSeq.len;                                  
      } /* endif */
   } /* endwhile */
   RxJisCountCheck(count, targetsize);
   target[count] = '\0';
   return (count);
}

static int
ToSbcs(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize = 0,
int linesize = 0) {
   RxJisArgCheck(source, target, targetsize, sourcesize);
   int count = 0;
   size_t sourcecount = 0; 
   while (sourcecount < sourcesize) { 
      if (!IsDbcs1st(source[sourcecount])) { 
         RxJisCountCheck(count, targetsize);
         target[count++] = source[sourcecount++]; 
      } else {
         int h = source[sourcecount++];
         int l = source[sourcecount++];
         if (!ToSbcsTable[h] || !ToSbcsTable[h][l]) {
            RxJisCountCheck(count + 1, targetsize);
            target[count++] = h;
            target[count++] = l;
         } else {
            for (int i = 0; ToSbcsTable[h][l][i]; i += 1) {
               RxJisCountCheck(count, targetsize);
               target[count++] = ToSbcsTable[h][l][i];
            } /* endfor */
         } /* endif */
      } /* endif */
   } /* endwhile */
   RxJisCountCheck(count, targetsize);
   target[count] = '\0';
   return (count);
}

static int
ToAscii(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize = 0,
int linesize = 0) {
   RxJisArgCheck(source, target, targetsize, sourcesize);
   int count = 0;
   size_t sourcecount = 0; 
   while (sourcecount < sourcesize) { 
      if (!IsDbcs1st(source[sourcecount])) { 
         RxJisCountCheck(count, targetsize);
         target[count++] = source[sourcecount++]; 
      } else {
         int h = source[sourcecount++];
         int l = source[sourcecount++];
         if (!ToSbcsTable[h] || !ToSbcsTable[h][l] || ToSbcsTable[h][l][0] > SCHAR_MAX) {
            RxJisCountCheck(count + 1, targetsize);
            target[count++] = h;
            target[count++] = l;
         } else {
            RxJisCountCheck(count, targetsize);
            target[count++] = ToSbcsTable[h][l][0];
         } /* endif */
      } /* endif */
   } /* endwhile */
   RxJisCountCheck(count, targetsize);
   target[count] = '\0';
   return (count);
}

static int
ToUpper(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize = 0,
int linesize = 0) {
   RxJisArgCheck(source, target, targetsize, sourcesize);
   int count = 0;
   wchar_t* sourcewide = new wchar_t[sourcesize + 1]; 
   if (!sourcewide) {
      return (CallingInvalid);
   } /* endif */
   const size_t sourcewidesize = nmbstowcs(sourcewide, sourcesize + 1, source, sourcesize); 
   size_t sourcewidecount = 0; 
   while (sourcewidecount < sourcewidesize) { 
      switch (sourcewide[sourcewidecount]) {
      case L'':
      case L'K':
      case L'':
      case L'J':
      case L'E':
      case L'':
      case L'[':
      case L'':
         sourcewidecount += 1;
         continue;
         break;
      } /* endswitch */
      if (sourcewide[sourcewidecount] <= UCHAR_MAX) { 
         RxJisCountCheck2(count, targetsize, sourcewide);
         target[count++] = UppercaseTable[0][sourcewide[sourcewidecount++]];
      } else {
         int h = sourcewide[sourcewidecount] >> ByteBitWidth; 
         int l = sourcewide[sourcewidecount] & ByteMask;      
         if (UppercaseTable[h]) {
            wchar_t c = UppercaseTable[h][l];
            h = c >> ByteBitWidth; 
            l = c & ByteMask;      
         } /* endif */
         RxJisCountCheck2(count + 1, targetsize, sourcewide);
         target[count++] = h;
         target[count++] = l;
         sourcewidecount += 1;
      } /* endif */
   } /* endwhile */
   RxJisCountCheck2(count, targetsize, sourcewide);
   target[count] = '\0';
   delete sourcewide; 
   return (count);
}

static int
ToDbcs(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize = 0,
int linesize = 0) {
   RxJisArgCheck(source, target, targetsize, sourcesize);
   int count = 0;
   wchar_t* sourcewide = new wchar_t[sourcesize + 1]; 
   if (!sourcewide) {
      return (CallingInvalid);
   } /* endif */
   const size_t sourcewidesize = nmbstowcs(sourcewide, sourcesize + 1, source, sourcesize); 
   size_t sourcewidecount = 0; 
   while (sourcewidecount < sourcewidesize) {
      wchar_t c = sourcewide[sourcewidecount++];
      if (c <= UCHAR_MAX) { 
         if (sourcewidecount >= sourcewidesize) {
            if (!ToDbcsTable1[c]) {
               RxJisCountCheck2(count, targetsize, sourcewide);
               target[count++] = c;
               continue;
            } /* endif */
            c = ToDbcsTable1[c];
         } else {
            wchar_t *t = ToDbcsTable1;
            switch (sourcewide[sourcewidecount]) {
            case L'':
            case L'K':
               t = ToDbcsTable3;
               sourcewidecount += 1;
               break;
            case L'':
            case L'J':
               t = ToDbcsTable2;
               sourcewidecount += 1;
               break;
            } /* endswitch */
            wchar_t d = t[c];
            if (!d) {
               d = ToDbcsTable1[c];
            } /* endif */
            if (d) {
               c = d;
            } /* endif */
            if (c <= UCHAR_MAX) {
               RxJisCountCheck2(count, targetsize, sourcewide);
               target[count++] = c;
               continue;
            } /* endif */
         } /* endif */
      } else if (sourcewidecount < sourcewidesize) {
         PWCHAR_T *t = NULL;
         switch (sourcewide[sourcewidecount]) {
            case L'':
            case L'J':
               t = ToDbcsTable4;
               sourcewidecount += 1;
               break;
         case L'':
         case L'K':
            t = ToDbcsTable5;
            sourcewidecount += 1;
            break;
         } /* endswitch */
         int h = c >> ByteBitWidth; 
         int l = c & ByteMask;      
         if (t && t[h]) {
            c = t[h][l];
         } /* endif */
      } /* endif */
      int h = c >> ByteBitWidth; 
      int l = c & ByteMask;      
      RxJisCountCheck2(count + 1, targetsize, sourcewide);
      target[count++] = h;
      target[count++] = l;
   } /* endwhile */
   RxJisCountCheck2(count, targetsize, sourcewide);
   target[count] = '\0';
   delete sourcewide; 
   return (count);
}

static int
ToZenkakuKatakana(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize = 0,
int linesize = 0) {
   RxJisArgCheck(source, target, targetsize, sourcesize);
   int count = 0;
   wchar_t* sourcewide = new wchar_t[sourcesize + 1]; 
   if (!sourcewide) {
      return (CallingInvalid);
   } /* endif */
   const size_t sourcewidesize = nmbstowcs(sourcewide, sourcesize + 1, source, sourcesize); 
   size_t sourcewidecount = 0; 
   while (sourcewidecount < sourcewidesize) {
      wchar_t c = sourcewide[sourcewidecount++];
      if (c <= SCHAR_MAX) {
         RxJisCountCheck2(count, targetsize, sourcewide);
         target[count++] = c;
         continue;
      } else if (c <= UCHAR_MAX) { 
         if (sourcewidecount >= sourcewidesize) {
            if (!ToDbcsTable1[c]) {
               RxJisCountCheck2(count, targetsize, sourcewide);
               target[count++] = c;
               continue;
            } /* endif */
            c = ToDbcsTable1[c];
         } else {
            wchar_t *t = ToDbcsTable1;
            switch (sourcewide[sourcewidecount]) {
            case L'':
            case L'K':
               t = ToDbcsTable3;
               sourcewidecount += 1;
               break;
            case L'':
            case L'J':
               t = ToDbcsTable2;
               sourcewidecount += 1;
               break;
            } /* endswitch */
            wchar_t d = t[c];
            if (!d) {
               d = ToDbcsTable1[c];
            } /* endif */
            if (d) {
               c = d;
            } /* endif */
            if (c <= UCHAR_MAX) {
               RxJisCountCheck2(count, targetsize, sourcewide);
               target[count++] = c;
               continue;
            } /* endif */
         } /* endif */
      } else if (sourcewidecount < sourcewidesize) {
         PWCHAR_T *t = NULL;
         switch (sourcewide[sourcewidecount]) {
            case L'':
            case L'J':
               t = ToDbcsTable4;
               sourcewidecount += 1;
               break;
         case L'':
         case L'K':
            t = ToDbcsTable5;
            sourcewidecount += 1;
            break;
         } /* endswitch */
         int h = c >> ByteBitWidth; 
         int l = c & ByteMask;      
         if (t && t[h]) {
            c = t[h][l];
         } /* endif */
      } /* endif */
      int h = c >> ByteBitWidth; 
      int l = c & ByteMask;      
      RxJisCountCheck2(count + 1, targetsize, sourcewide);
      target[count++] = h;
      target[count++] = l;
   } /* endwhile */
   RxJisCountCheck2(count, targetsize, sourcewide);
   target[count] = '\0';
   delete sourcewide; 
   return (count);
}

static int
ToQuotedPrintable(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize = 0,
int linesize = 0) {
   RxJisArgCheck(source, target, targetsize, sourcesize);
   int count = 0;
   int column = 0;
   for (int i = 1; i <= sourcesize; i += 1) {
      if (count + ToQuotedPrintableTable[*source].len >= targetsize) {
         return (RxJisError); 
      } /* endif */
      if (column + ToQuotedPrintableTable[*source].len > MaxMimeLineSize) {
         if (count + 3 >= targetsize) {
            return (RxJisError); 
         } /* endif */
         target[count++] = '=';
         target[count++] = '\r';
         target[count++] = '\n';
         column = 0;
      } /* endif */
      strncpy(target + count, ToQuotedPrintableTable[*source].seq, ToQuotedPrintableTable[*source].len);
      count += ToQuotedPrintableTable[*source].len;
      column += ToQuotedPrintableTable[*source].len;
      source += 1;
   } /* endfor */
   if (count >= targetsize) {
      return (RxJisError); 
   } /* endif */
   target[count] = '\0';
   return (count);
}

static int
ToBase64(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize = 0,
int linesize = 0) {
   RxJisArgCheck(source, target, targetsize, sourcesize);
   int count = 0;
   int column = 0;
   for (int i = 1; i <= sourcesize; i += 3) {
      RxJisCountCheck(count + 3, targetsize);
      if (column >= MaxMimeLineSize) {
         RxJisCountCheck(count + 2, targetsize);
         target[count++] = '\r';
         target[count++] = '\n';
         column = 0;
      } /* endif */
      long v;
      if (i + 2 <= sourcesize) {
         v = *source++ << (2 * ByteBitWidth);
         v |= *source++ << ByteBitWidth;
         v |= *source++;
         target[count++] = ToBase64Table[Base64Mask&(v>>(3*Base64BitWidth))];
         target[count++] = ToBase64Table[Base64Mask&(v>>2*Base64BitWidth)];
         target[count++] = ToBase64Table[Base64Mask&(v>>Base64BitWidth)];
         target[count++] = ToBase64Table[Base64Mask&v];
      } else if (i + 1 <= sourcesize) {
         v = *source++ << ByteBitWidth;
         v |= *source++;
         v <<= (3*Base64BitWidth-2*ByteBitWidth);
         target[count++] = ToBase64Table[Base64Mask&(v>>2*Base64BitWidth)];
         target[count++] = ToBase64Table[Base64Mask&(v>>Base64BitWidth)];
         target[count++] = ToBase64Table[Base64Mask&v];
         target[count++] = '=';
      } else {
         v = *source++ << (2*Base64BitWidth-ByteBitWidth);
         target[count++] = ToBase64Table[Base64Mask&(v>>Base64BitWidth)];
         target[count++] = ToBase64Table[Base64Mask&v];
         target[count++] = '=';
         target[count++] = '=';
      } /* endif */
   } /* endfor */
   RxJisCountCheck(count, targetsize);
   target[count] = '\0';
   return (count);
}

static int
Iso8859_1ToCp850(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize,  
int linesize) {  
   RxJisArgCheck(source, target, targetsize, sourcesize);
   RxJisCountCheck(sourcesize, targetsize);
   int count = 0;
   for (int sourcecount = 0; sourcecount < sourcesize; sourcecount += 1) {
      RxJisCountCheck(count, targetsize);
      target[count++] = ToCp850(source[sourcecount]);
   } /* endfor */
   RxJisCountCheck(count, targetsize);
   target[count] = '\0';
   return (count);
}

static int
Cp850ToIso8859_1(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize,  
int linesize) {  
   RxJisArgCheck(source, target, targetsize, sourcesize);
   RxJisCountCheck(sourcesize, targetsize);
   int count = 0;
   for (int sourcecount = 0; sourcecount < sourcesize; sourcecount += 1) {
      RxJisCountCheck(count, targetsize);
      target[count++] = ToIso8859_1(source[sourcecount]);
   } /* endfor */
   RxJisCountCheck(count, targetsize);
   target[count] = '\0';
   return (count);
}

static int
Cp852ToIso8859_2(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize,  
int linesize) {  
   RxJisArgCheck(source, target, targetsize, sourcesize);
   RxJisCountCheck(sourcesize, targetsize);
   int count = 0;
   for (int sourcecount = 0; sourcecount < sourcesize; sourcecount += 1) {
      RxJisCountCheck(count, targetsize);
      target[count++] = ToIso8859_2(source[sourcecount]);
   } /* endfor */
   RxJisCountCheck(count, targetsize);
   target[count] = '\0';
   return (count);
}

static int
Cp857ToIso8859_9(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize,  
int linesize) {  
   RxJisArgCheck(source, target, targetsize, sourcesize);
   RxJisCountCheck(sourcesize, targetsize);
   int count = 0;
   for (int sourcecount = 0; sourcecount < sourcesize; sourcecount += 1) {
      RxJisCountCheck(count, targetsize);
      target[count++] = ToIso8859_9(source[sourcecount]);
   } /* endfor */
   RxJisCountCheck(count, targetsize);
   target[count] = '\0';
   return (count);
}

static int
Iso8859_2ToCp852(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize,  
int linesize) {  
   RxJisArgCheck(source, target, targetsize, sourcesize);
   RxJisCountCheck(sourcesize, targetsize);
   int count = 0;
   for (int sourcecount = 0; sourcecount < sourcesize; sourcecount += 1) {
      target[count++] = ToCp852(source[sourcecount]);
   } /* endfor */
   RxJisCountCheck(count, targetsize);
   target[count] = '\0';
   return (count);
}

static int
Iso8859_9ToCp857(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize,  
int linesize) {  
   RxJisArgCheck(source, target, targetsize, sourcesize);
   RxJisCountCheck(sourcesize, targetsize);
   int count = 0;
   for (int sourcecount = 0; sourcecount < sourcesize; sourcecount += 1) {
      target[count++] = ToCp857(source[sourcecount]);
   } /* endfor */
   RxJisCountCheck(count, targetsize);
   target[count] = '\0';
   return (count);
}

static int
ToPc(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize,  
int linesize) {  
   RxJisArgCheck(source, target, targetsize, sourcesize);
   int count = 0;
   int sourcecount = 0;
   CharSet currentset = USASCII; 
   CHAR currentmask = 0x00;
   PWCHAR_T* jistopccode; 
   while (sourcecount < sourcesize) { 
      int i;
      switch (source[sourcecount]) {
      case SO:
      case SI:
         target[count++] = source[sourcecount++];
         break;
      case Esc:
         for (i = 0; DoubleByteSeqTable[i].seq; i += 1) {
            if (sourcecount + DoubleByteSeqTable[i].len > sourcesize) {
               continue; 
            } /* endif */
            if (strncmp(source + sourcecount, DoubleByteSeqTable[i].seq, DoubleByteSeqTable[i].len) == 0) {
               break;
            } /* endif */
         } /* endfor */
         if (DoubleByteSeqTable[i].seq != NULL) {
            if (sourcecount + DoubleByteSeqTable[i].len > sourcesize) {
               RxJisCountCheck(count, targetsize);
               target[count++] = source[sourcecount++];
               continue;
            } /* endif */
            sourcecount += DoubleByteSeqTable[i].len;
            switch (DoubleByteSeqTable[i].charset) { 
            case USASCII:
               switch (currentset) {
               case ISO8859_1:
               case ISO8859_7:
                  RxJisCountCheck(count + Jis83OutSeq.len, targetsize);
                  strncpy(target + count, Jis83OutSeq.seq, Jis83OutSeq.len);
                  count += Jis83OutSeq.len;
                  break;
               default:
                 break;
               } /* endswitch */
               currentset = DoubleByteSeqTable[i].charset;
               break;
            case SS2:
               if (sourcecount < sourcesize) {
                  RxJisCountCheck(count, targetsize);
                  switch (currentset) {
                  case ISO8859_1:
                     target[count++] = ToCp850(ByteMsb | source[sourcecount++]); 
                     break;
                  default:
                    target[count++] = source[sourcecount++];                 
                    break;
                  } /* endswitch */
               } /* endif */
               break;
            case ISO8859_1:
            case ISO8859_7:
               RxJisCountCheck(count + DoubleByteSeqTable[i].len, targetsize);
               strncpy(target + count, DoubleByteSeqTable[i].seq, DoubleByteSeqTable[i].len);
               currentset = DoubleByteSeqTable[i].charset;
               break;
            default:
              switch (currentset) {
              case ISO8859_1:
              case ISO8859_7:
              case ISO2022KR:
                 RxJisCountCheck(count + Jis83OutSeq.len, targetsize);
                 strncpy(target + count, Jis83OutSeq.seq, Jis83OutSeq.len);
                 count += Jis83OutSeq.len;
                 break;
              } /* endswitch */
              currentset = DoubleByteSeqTable[i].charset;
              jistopccode = DoubleByteSeqTable[i].jistopccodetable;
              break;
            } /* endswitch */
         } else {
            RxJisCountCheck(count, targetsize);
            currentset = USASCII;
            target[count++] = source[sourcecount++];                               
         } /* endif */
         break;
      default: 
        wchar_t c;
        int l;
        int h;
        switch (currentset) {
        case HankakuKatakana:
           RxJisCountCheck(count, targetsize);
           target[count++] = ByteMsb | source[sourcecount++];
           break;
        case ISO2022JP:
        case JISX0212_1990:
        case ISO2022KR:
           if (sourcecount + 2 > sourcesize) { 
              RxJisCountCheck(count, targetsize);
              target[count++] = source[sourcecount++];
              continue;
           } /* endif */                                     
           RxJisCountCheck(count + 2, targetsize);
           if (source[sourcecount] <= ' ' || source[sourcecount] <= ' ') { 
              target[count++] = source[sourcecount++];       
              target[count++] = source[sourcecount++];       
              currentset = USASCII;
              continue;                                      
           } /* endif */                                     
           h = source[sourcecount++];           
           l = source[sourcecount++];           
           c = JisToPc(h << ByteBitWidth | l, jistopccode);              
           target[count++] = c >> ByteBitWidth;      
           target[count++] = c & ByteMask;           
           break;
        default:
          RxJisCountCheck(count, targetsize);
          target[count++] = source[sourcecount++];
          break;
        } /* endswitch */
        break;
      } /* endswitch */
   } /* endwhile */
   switch (currentset) {
   case ISO8859_1:
   case ISO8859_7:
   case ISO2022KR:
      RxJisCountCheck(count + Jis83OutSeq.len, targetsize);
      strncpy(target + count, Jis83OutSeq.seq, Jis83OutSeq.len);
      count += Jis83OutSeq.len;
      break;
   } /* endswitch */
   RxJisCountCheck(count, targetsize);
   target[count] = '\0';
   return (count);
}

static int
QuotedPrintableTo(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize, 
int linesize) { 
   RxJisArgCheck(source, target, targetsize, sourcesize);
   int count = 0;
   int sourcecount = 0; 
   while (sourcecount < sourcesize) { 
      long v;
      long c;
      int p;
      switch (source[sourcecount]) {
      case '=':
         RxJisCountCheck(count, targetsize);
         if (!source[++sourcecount]) {
            continue;
         } /* endif */
         v = QuotedPrintableTable[source[sourcecount]];
         if (QuotedPrintableError == v) {
            continue;
         } /* endif */
         if (!source[++sourcecount]) {
            continue;
         } /* endif */
         c = QuotedPrintableTable[source[sourcecount++]];
         if (QuotedPrintableError == c) {
            continue;
         } /* endif */
         target[count++] = (v << 4) | c;
         break;
      case Space:
      case Ht:
         p = sourcecount;
         while (source[p]) {
            if (source[p] != Space && source[p] != Ht) {
               break;
            } /* endif */
            p += 1;
         } /* endwhile */
         if (source[p]) {
            p = sourcecount;
            while (source[p]) {
               RxJisCountCheck(count, targetsize);
               if (source[p] != Space && source[p] != Ht) {
                  break;
               } /* endif */
               target[count++] = source[p++];
            } /* endwhile */
         } /* endif */
         sourcecount = p;
         break;
      default: 
         RxJisCountCheck(count, targetsize);
         target[count++] = source[sourcecount++];
         break;
      } /* endswitch */
   } /* endwhile */
   RxJisCountCheck(count, targetsize);
   target[count] = '\0';
   return (count);
}

static int
Base64To(
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize, 
int linesize) { 
   RxJisArgCheck(source, target, targetsize, sourcesize);
   int count = 0;
   for (int i = 1; i <= sourcesize; i += 4) {
      long v = 0;
      int m = 0;
      for (int j = 1; *source && j <= 4; j += 1) {
         long c = Base64Table[*source++];
         switch (c) {
         case Base64Error:
         case Base64Pad:
            break;
         default:
            v = (v << Base64BitWidth) | c;
            m += Base64BitWidth;
            break;
         } /* endswitch */
      } /* endfor */
      if (m >= ByteBitWidth) {
         v >>= (m % ByteBitWidth);
         m /= ByteBitWidth;
      } else {
         v <<= (ByteBitWidth - m);
         m = 1;
      } /* endif */
      if (count + m > targetsize) {
         return (RxJisError); 
      } /* endif */
      for (j = m - 1; j >= 0; j -= 1) {
         target[count+j] = v & ByteMask;
         v >>= ByteBitWidth;
      } /* endfor */
      count += m;
   } /* endfor */
   target[count] = '\0';
   return (count);
}

static ULONG
JisCommon(
int (*convert)(PCHAR, PCHAR, int, int, int),
PCHAR target,
PCHAR source,
int targetsize,
int sourcesize = 0,
int (*mintargetsize)(PCHAR, int, int) = NULL) {
   if (!TableIsLoaded) {  
      LoadRxJisTables();  
   } /* endif */          
   if (0 == sourcesize) {
      sourcesize = strlen(source);
   } /* endif */
   if (0 == sourcesize) {
      target[0] = '\0';
      return (0);
   } /* endif */
   int minsize;
   if (mintargetsize) {
      minsize = (*mintargetsize)(source, sourcesize, 0);
   } else {
      minsize = sourcesize;
   } /* endif */
   if (targetsize < minsize) {
      return (RxJisError);
   } /* endif */
   return ((*convert)(target, source, targetsize, sourcesize, 0));
}

int _Export 
JisToJis(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(ToJis, target, source, targetsize, sourcesize, JisSize));
}

int _Export 
JisToEuc(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(ToEuc, target, source, targetsize, sourcesize, EucSize));
}

int _Export 
JisToBase64(
char* target,
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(ToBase64, target, source, targetsize, sourcesize, Base64Size));
}

int _Export 
JisToQuotedPrintable(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(ToQuotedPrintable, target, source, targetsize, sourcesize, QuotedPrintableSize));
}

int _Export JisJisTo(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(ToPc, target, source, targetsize, sourcesize));
}

int _Export 
JisBase64To(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(Base64To, target, source, targetsize, sourcesize));
}

int _Export 
JisEucTo(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(EucTo, target, source, targetsize, sourcesize));
}

int _Export 
JisQuotedPrintableTo(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(QuotedPrintableTo, target, source, targetsize, sourcesize));
}

int _Export 
JisMimeJisTo(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(MimeJisTo, target, source, targetsize, sourcesize));
}

int _Export 
JisJisCharTo(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(JisCharTo, target, source, targetsize, sourcesize));
}

int _Export 
JisToJisChar(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(ToJisChar, target, source, targetsize, sourcesize));
}

int _Export 
JisToSbcs(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(ToSbcs, target, source, targetsize, sourcesize));
}

int _Export 
JisToDbcs(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(ToDbcs, target, source, targetsize, sourcesize, DbcsSize));
}

int _Export 
JisToAscii(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(ToAscii, target, source, targetsize, sourcesize));
}

int _Export 
JisToUpper(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(ToUpper, target, source, targetsize, sourcesize));
}

int _Export 
JisCp850ToIso8859_1(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(Cp850ToIso8859_1, target, source, targetsize, sourcesize));
}

int _Export 
JisCp852ToIso8859_2(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(Cp852ToIso8859_2, target, source, targetsize, sourcesize));
}

int _Export 
JisCp857ToIso8859_9(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(Cp857ToIso8859_9, target, source, targetsize, sourcesize));
}

int _Export 
JisIso8859_9ToCp857(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(Iso8859_9ToCp857, target, source, targetsize, sourcesize));
}

int _Export 
JisIso8859_2ToCp852(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(Iso8859_2ToCp852, target, source, targetsize, sourcesize));
}

int _Export 
JisIso8859_1ToCp850(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(Iso8859_1ToCp850, target, source, targetsize, sourcesize));
}

int _Export 
JisToZenkakuKatakana(
char* target, 
char* source, 
int targetsize, 
int sourcesize) {
   return (JisCommon(ToZenkakuKatakana, target, source, targetsize, sourcesize));
}

static ULONG
RxJisCommon(
PUCHAR name, 
ULONG argc,
RXSTRING argv[],
PSZ queuename,
RXSTRING *retstr,
int (*convert)(PCHAR, PCHAR, int, int, int),
int (*mintargetsize)(PCHAR, int, int) = NULL) {
   if (0 == argc || argc > 2) {
      return (CallingInvalid);
   } /* endif */
   if (NULL == retstr || RXNULLSTRING(argv[0])) {
      return (CallingInvalid);
   } /* endif */
   PSZ& source = RXSTRPTR(argv[0]);
   const int& sourcesize = RXSTRLEN(argv[0]);
   int linesize = 0;
   if (argc > 1) {
      if (RXNULLSTRING(argv[1])) {
         return (CallingInvalid);
      } /* endif */
      int n = sscanf(RXSTRPTR(argv[1]), "%d", &linesize);
      if (n != 1 || linesize < MaxMimeLineSize / 2 || linesize > MaxMimeLineSize) {
         return (CallingInvalid);
      } /* endif */
   } /* endif */
   if (0 == sourcesize) {    
      retstr->strlength = 0; 
      return (CallingValid); 
   } /* endif */             
   if (!TableIsLoaded) {  
      LoadRxJisTables();  
   } /* endif */          
   int minimumtargetsize;
   if (mintargetsize) {                                                   
      minimumtargetsize = (*mintargetsize)(source, sourcesize, linesize); 
   } else {                                                               
      minimumtargetsize = sourcesize + 1;                                 
   } /* endif */                                                          
   if (RXNULLSTRING(*retstr) || RXSTRLEN(*retstr) < minimumtargetsize) {  
      retstr->strlength = minimumtargetsize; 
      APIRET code = DosAllocMem((PPVOID)&retstr->strptr, retstr->strlength, PAG_COMMIT | PAG_READ | PAG_WRITE);
      if (code != NO_ERROR) {
         return (CallingInvalid);
      } /* endif */
   } /* endif */
   int count = (*convert)(RXSTRPTR(*retstr), source, RXSTRLEN(*retstr), sourcesize, linesize);
   if (count < 0) {
      return (CallingInvalid);
   } /* endif */
   retstr->strlength = count;
   return (CallingValid);
}

ULONG _Export 
RxJisBase64To(
PUCHAR name, 
ULONG argc,
RXSTRING argv[],
PSZ queuename,
RXSTRING *retstr) {
   return (RxJisCommon(name, argc, argv, queuename, retstr, Base64To));
}

ULONG _Export 
RxJisToPc(
PUCHAR name, 
ULONG argc,
RXSTRING argv[],
PSZ queuename,
RXSTRING *retstr) {
   return (RxJisCommon(name, argc, argv, queuename, retstr, ToPc));
}

ULONG _Export
RxJisJisTo(
PUCHAR name, 
ULONG argc,
RXSTRING argv[],
PSZ queuename,
RXSTRING *retstr) {
   return (RxJisCommon(name, argc, argv, queuename, retstr, ToPc));
}

ULONG _Export
RxJisToZenkakuKatakana(
PUCHAR name, 
ULONG argc,
RXSTRING argv[],
PSZ queuename,
RXSTRING *retstr) {
   return (RxJisCommon(name, argc, argv, queuename, retstr, ToZenkakuKatakana));
}

ULONG _Export 
RxJisQuotedPrintableTo(
PUCHAR name, 
ULONG argc,
RXSTRING argv[],
PSZ queuename,
RXSTRING *retstr) {
   return (RxJisCommon(name, argc, argv, queuename, retstr, QuotedPrintableTo));
}

ULONG _Export 
RxJisToJis(
PUCHAR name, 
ULONG argc,
RXSTRING argv[],
PSZ queuename,
RXSTRING *retstr) {
   return (RxJisCommon(name, argc, argv, queuename, retstr, ToJis, JisSize));
}

ULONG _Export 
RxJisToBase64(
PUCHAR name, 
ULONG argc,
RXSTRING argv[],
PSZ queuename,
RXSTRING *retstr) {
   return (RxJisCommon(name, argc, argv, queuename, retstr, ToBase64, Base64Size));
}

ULONG _Export 
RxJisToQuotedPrintable(
PUCHAR name, 
ULONG argc,
RXSTRING argv[],
PSZ queuename,
RXSTRING *retstr) {
   return (RxJisCommon(name, argc, argv, queuename, retstr, ToQuotedPrintable, QuotedPrintableSize));
}

ULONG _Export                                                                
RxJisToEuc(                                                                  
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   return (RxJisCommon(name, argc, argv, queuename, retstr, ToEuc, EucSize));
}                                                                            

ULONG _Export                                                                
RxJisEucTo(                                                                  
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   return (RxJisCommon(name, argc, argv, queuename, retstr, EucTo));
}                                                                            

ULONG _Export                                                                
RxJisMimeJisTo(                                                              
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   return (RxJisCommon(name, argc, argv, queuename, retstr, MimeJisTo));
}                                                                            

ULONG _Export                                                                
RxJisJisCharTo(                                                              
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   return (RxJisCommon(name, argc, argv, queuename, retstr, JisCharTo));
}                                                                            

ULONG _Export                                                                
RxJisToJisChar(                                                              
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   return (RxJisCommon(name, argc, argv, queuename, retstr, ToJisChar));
}                                                                            

ULONG _Export                                                                
RxJisToDbcs(                                                              
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   return (RxJisCommon(name, argc, argv, queuename, retstr, ToDbcs, DbcsSize));
}                                                                            

ULONG _Export                                                                
RxJisToSbcs(                                                              
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   return (RxJisCommon(name, argc, argv, queuename, retstr, ToSbcs));
}                                                                            

ULONG _Export                                                                
RxJisToAscii(                                                              
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   return (RxJisCommon(name, argc, argv, queuename, retstr, ToAscii));
}                                                                            

ULONG _Export                                                                
RxJisToUpper(                                                              
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   return (RxJisCommon(name, argc, argv, queuename, retstr, ToUpper));
}                                                                            

ULONG _Export                                                                
RxJisIso8859_1ToCp850(                                                              
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   return (RxJisCommon(name, argc, argv, queuename, retstr, Iso8859_1ToCp850));
}                                                                            

ULONG _Export                                                                
RxJisIso8859_2ToCp852(                                                              
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   return (RxJisCommon(name, argc, argv, queuename, retstr, Iso8859_2ToCp852));
}                                                                            

ULONG _Export                                                                
RxJisIso8859_9ToCp857(                                                              
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   return (RxJisCommon(name, argc, argv, queuename, retstr, Iso8859_9ToCp857));
}                                                                            

ULONG _Export                                                                
RxJisCp857ToIso8859_9(                                                              
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   return (RxJisCommon(name, argc, argv, queuename, retstr, Cp857ToIso8859_9));
}                                                                            

ULONG _Export                                                                
RxJisCp852ToIso8859_2(                                                              
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   return (RxJisCommon(name, argc, argv, queuename, retstr, Cp852ToIso8859_2));
}                                                                            

ULONG _Export                                                                
RxJisCp850ToIso8859_1(                                                              
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   return (RxJisCommon(name, argc, argv, queuename, retstr, Cp850ToIso8859_1));
}                                                                            

ULONG _Export                                                                
RxJisWords(                                                              
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   if (argc != 1) {
      return (CallingInvalid);
   } /* endif */
   int count = JisWords(RXSTRPTR(argv[0]), RXSTRLEN(argv[0]));
   if (count < 0) {
      return (CallingInvalid);
   } /* endif */
   sprintf(RXSTRPTR(*retstr), "%d", count);
   retstr->strlength = strlen(RXSTRPTR(*retstr));
   return (CallingValid);
}                                                                            

ULONG _Export                                                                
RxJisWord(                                                              
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   if (argc != 2) {
      return (CallingInvalid);
   } /* endif */
   if (NULL == retstr || RXNULLSTRING(argv[0]) || RXNULLSTRING(argv[1])) {
      return (CallingInvalid);
   } /* endif */
   PSZ& source = RXSTRPTR(argv[0]);
   const int& sourcesize = RXSTRLEN(argv[0]);
   int position = 0;
   int n = sscanf(RXSTRPTR(argv[1]), "%d", &position);
   if (n != 1 || position < 1) {
      return (CallingInvalid);
   } /* endif */
   if (0 == sourcesize) {    
      retstr->strlength = 0; 
      return (CallingValid); 
   } /* endif */             
   if (!TableIsLoaded) {  
      LoadRxJisTables();  
   } /* endif */          
   int minimumtargetsize = sourcesize + 1;                                 
   if (RXNULLSTRING(*retstr) || RXSTRLEN(*retstr) < minimumtargetsize) {  
      retstr->strlength = minimumtargetsize; 
      APIRET code = DosAllocMem((PPVOID)&retstr->strptr, retstr->strlength, PAG_COMMIT | PAG_READ | PAG_WRITE);
      if (code != NO_ERROR) {
         return (CallingInvalid);
      } /* endif */
   } /* endif */
   PSZ& target = RXSTRPTR(*retstr);
   const int& targetsize = RXSTRLEN(*retstr);
   int count = JisWord(target, source, targetsize, position, sourcesize);
   if (count < 0) {
      return (CallingInvalid);
   } /* endif */
   retstr->strlength = count;
   return (CallingValid);
}                                                                            

ULONG _Export                                                                
RxJisAppend(                                                              
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   if (argc != 2) {
      return (CallingInvalid);
   } /* endif */
   if (NULL == retstr || RXNULLSTRING(argv[0]) || RXNULLSTRING(argv[1])) {
      return (CallingInvalid);
   } /* endif */
   if (!TableIsLoaded) {  
      LoadRxJisTables();  
   } /* endif */          
   PSZ& source = RXSTRPTR(argv[1]);
   const int& sourcesize = RXSTRLEN(argv[1]);
   const int& targetlength = RXSTRLEN(argv[0]);
   const int minimumtargetsize = targetlength + 1 + sourcesize + 1;                                 
   PSZ& target = RXSTRPTR(*retstr);
   if (!target || RXSTRLEN(*retstr) < minimumtargetsize) {  
      retstr->strlength = minimumtargetsize; 
      APIRET code = DosAllocMem((PPVOID)&target, RXSTRLEN(*retstr), PAG_COMMIT | PAG_READ | PAG_WRITE);
      if (code != NO_ERROR) {
         return (CallingInvalid);
      } /* endif */
   } /* endif */
   const int& targetsize = RXSTRLEN(*retstr);
   memcpy(target, RXSTRPTR(argv[0]), RXSTRLEN(argv[0]) + 1);
   int count = JisAppend(target, source, targetsize, targetlength, sourcesize);
   if (count < 0) {
      return (CallingInvalid);
   } /* endif */
   retstr->strlength = count;
   return (CallingValid);
}                                                                            

static int
parsenamevalue(
PCHAR target,
int targetsize,
PCHAR source,
int& sourceposition,
int sourcesize,
CHAR delimiter) {
   int count = 0;
   while (sourceposition < sourcesize) {
      if (IsDbcs1st(source[sourceposition])) {
         target[count++] = source[sourceposition++];
         if (sourceposition < sourcesize) {
            target[count++] = source[sourceposition++];
         } /* endif */
         continue;
      } /* endif */
      if (delimiter == source[sourceposition]) {
         sourceposition += 1;
         break;
      } /* endif */
      switch (source[sourceposition]) {
      case '%':
         sourceposition += 1;
         if (sourceposition + 1 < sourcesize) {
            int code;
            sscanf(&source[sourceposition], "%2x", &code);
            sourceposition += 2;
            target[count++] = code;
         } /* endif */
         break;
      case '+':
         target[count++] = ' ';
         sourceposition += 1;
         break;
      default: 
        target[count++] = source[sourceposition++];
        break;
      } /* endswitch */
   } /* endwhile */
   target[count] = '\0';
   return (count);
}

static int
setrexxvar(
PCHAR varname,
PCHAR varvalue,
int varvaluesize) {
   SHVBLOCK block;
   block.shvcode = RXSHV_SYSET;
   block.shvret = RXSHV_OK;
   block.shvnext = NULL;
   MAKERXSTRING(block.shvname, varname, strlen(varname));
   MAKERXSTRING(block.shvvalue, varvalue, varvaluesize);
   ULONG rc = RexxVariablePool(&block);
   if (RXSHV_NOAVL == rc || ~(RXSHV_OK | RXSHV_NEWV) & block.shvret) {
      return (CallingInvalid);
   } /* endif */
   return (CallingValid);
}

ULONG _Export                                                                
RxJisUrlToVar(                                                              
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   if (argc < 1 || argc > 2) {
      return (CallingInvalid);
   } /* endif */
   if (NULL == retstr || RXNULLSTRING(argv[0]) || (2 == argc && RXNULLSTRING(argv[1]) || RXSTRLEN(argv[1]) < 1)) {
      return (CallingInvalid);
   } /* endif */
   if (!TableIsLoaded) {  
      LoadRxJisTables();  
   } /* endif */          
   PSZ& url = RXSTRPTR(argv[0]);
   const int& urlsize = RXSTRLEN(argv[0]);
   const int worksize = urlsize + 1;
   PSZ stemname;
   if (2 == argc) {
      stemname = new CHAR[RXSTRLEN(argv[1]) + 1];
      strncpy(stemname, RXSTRPTR(argv[1]), RXSTRLEN(argv[1]) + 1);
      if ('.' == stemname[RXSTRLEN(argv[1]) - 1]) {
         stemname[RXSTRLEN(argv[1]) - 1] = '\0';
      } /* endif */
   } else {
      stemname = new CHAR[4];
      strncpy(stemname, "URL", 4);
   } /* endif */
   PCHAR varname = new CHAR[worksize];
   PCHAR varvalue = new CHAR[worksize];
   PCHAR temp = new CHAR[worksize];
   sprintf(varname, "%s.", stemname);
   if (setrexxvar(varname, "", 0) != CallingValid) {
      return (CallingInvalid);
   } /* endif */
   int i = 0;
   parsenamevalue(temp, worksize, url, i, urlsize, '?');
   while (i < urlsize) {
      int tempsize = parsenamevalue(temp, worksize, url, i, urlsize, '=');
      sprintf(varname, "%s.%s", stemname, temp);
      tempsize = parsenamevalue(temp, worksize, url, i, urlsize, '&');
      int varvaluesize = JisJisTo(varvalue, temp, worksize, tempsize);
      if (setrexxvar(varname, varvalue, varvaluesize) != CallingValid) {
         return (CallingInvalid);
      } /* endif */
   } /* endwhile */
   delete stemname;
   delete varname;
   delete varvalue;
   delete temp;
   retstr->strlength = 0;
   return (CallingValid);
}                                                                            

ULONG _Export                                                                
RxJisWordIndex(
PUCHAR name,                                                                 
ULONG argc,                                                                  
RXSTRING argv[],                                                             
PSZ queuename,                                                               
RXSTRING *retstr) {                                                          
   if (argc != 2) {
      return (CallingInvalid);
   } /* endif */
   if (NULL == retstr || RXNULLSTRING(argv[0]) || RXNULLSTRING(argv[1])) {
      return (CallingInvalid);
   } /* endif */
   PSZ& source = RXSTRPTR(argv[0]);
   const int& sourcesize = RXSTRLEN(argv[0]);
   int position = 0;
   int n = sscanf(RXSTRPTR(argv[1]), "%d", &position);
   if (n != 1 || position < 1) {
      return (CallingInvalid);
   } /* endif */
   if (0 == sourcesize) {    
      retstr->strlength = 0; 
      return (CallingValid); 
   } /* endif */             
   if (!TableIsLoaded) {  
      LoadRxJisTables();  
   } /* endif */          
   int minimumtargetsize = sourcesize + 1;                                 
   if (RXNULLSTRING(*retstr) || RXSTRLEN(*retstr) < minimumtargetsize) {  
      retstr->strlength = minimumtargetsize; 
      APIRET code = DosAllocMem((PPVOID)&retstr->strptr, retstr->strlength, PAG_COMMIT | PAG_READ | PAG_WRITE);
      if (code != NO_ERROR) {
         return (CallingInvalid);
      } /* endif */
   } /* endif */
   PSZ& target = RXSTRPTR(*retstr);
   const int& targetsize = RXSTRLEN(*retstr);
   int count = JisWordIndex(target, source, targetsize, position, sourcesize);
   if (count < 0) {
      return (CallingInvalid);
   } /* endif */
   sprintf(RXSTRPTR(*retstr), "%d", count);
   retstr->strlength = strlen(RXSTRPTR(*retstr));
   return (CallingValid);
}                                                                            

static wchar_t
JisCodeToPcCode(
wchar_t jiscode) {
   int h = jiscode >> ByteBitWidth;
   int l = jiscode & ByteMask;
   if (h & '\x01') {
      l += '\x1f';
   } else {
      l += '\x7d';
   } /* endif */
   if (l >= '\x7f') {
      l += 1;
   } /* endif */
   h = (h - '\x21') >> 1;
   if (h > '\x1e') {
      h += '\xc1';
   } else {
      h += '\x81';
   } /* endif */
   return (h << ByteBitWidth | l);
}

static wchar_t
PcCodeToJisCode(
wchar_t pccode) {
   int h = pccode >> ByteBitWidth;
   int l = pccode & ByteMask;
   if (h <= '\x9f') {
      h -= '\x71';
   } else {
      h -= '\xb1';
   } /* endif */
   h = 2 * h + 1;
   if (l > '\x7f') {
      l -= 1;
   } /* endif */
   if (l >= '\x9e') {
      l -= '\x7d';
      h += 1;
   } else {
      l -= '\x1f';
   } /* endif */
   return (h << ByteBitWidth | l);
}

static wchar_t
KutenCodeToJisCode(
wchar_t kutencode) {
   int h = kutencode / 100 + 0x20;
   int l = kutencode % 100 + 0x20;
   return (h << ByteBitWidth | l);
}

static APIRET
LoadRxJisTables(void) {
   if (TableIsLoaded) {  
      return (NO_ERROR); 
   } /* endif */         
   for (int i = 0; i <= UCHAR_MAX; i += 1) {
      QuotedPrintableTable[i] = QuotedPrintableError;
   } /* endfor */
   for (i = '0'; i <= '9'; i += 1) {
      QuotedPrintableTable[i] = i - '0';
   } /* endfor */
   for (i = 'A'; i <= 'F'; i += 1) {
      QuotedPrintableTable[i] = 0x0a + i - 'A';
   } /* endfor */
   for (i = 'a'; i <= 'f'; i += 1) {
      QuotedPrintableTable[i] = 0x0a + i - 'a';
   } /* endfor */
   for (i = 0; i < Space; i += 1) {
      ToQuotedPrintableTable[i].seq = new CHAR[4];
      if (!ToQuotedPrintableTable[i].seq) {
         return (ERROR_NOT_ENOUGH_MEMORY);
      } /* endif */
      sprintf(ToQuotedPrintableTable[i].seq, "=%02X", i);
      ToQuotedPrintableTable[i].len = 3;
   } /* endfor */
   for (i = Space; i <= 126; i += 1) {
      ToQuotedPrintableTable[i].seq = new CHAR[4];
      if (!ToQuotedPrintableTable[i].seq) {
         return (ERROR_NOT_ENOUGH_MEMORY);
      } /* endif */
      sprintf(ToQuotedPrintableTable[i].seq, "%c", i);
      ToQuotedPrintableTable[i].len = 1;
   } /* endfor */
   for (i = 127; i <= UCHAR_MAX; i += 1) {
      ToQuotedPrintableTable[i].seq = new CHAR[4];
      if (!ToQuotedPrintableTable[i].seq) {
         return (ERROR_NOT_ENOUGH_MEMORY);
      } /* endif */
      sprintf(ToQuotedPrintableTable[i].seq, "=%02X", i);
      ToQuotedPrintableTable[i].len = 3;
   } /* endfor */
   strcpy(ToQuotedPrintableTable['\t'].seq, "\t");
   ToQuotedPrintableTable['\t'].len = 1;
   strcpy(ToQuotedPrintableTable['\x1b'].seq, "\x1b");
   ToQuotedPrintableTable['\x1b'].len = 1;
   strcpy(ToQuotedPrintableTable['='].seq, "=3D");
   ToQuotedPrintableTable['='].len = 3;
   ULONG cplistsize;
   ULONG cplist[CpListSize];
   APIRET rc = DosQueryCp(sizeof(cplist), cplist, &cplistsize);
   if (rc != NO_ERROR) {
      return (rc);
   } /* endif */
   DefaultCodepage = cplist[0];
   if (943 == DefaultCodepage) {
      InJis90 = TRUE;
   } else {
      PCHAR v = getenv(JisEnvVarName);                                                  
      if (v) {                                                                          
         PCHAR s = strstr(v, Jis90Value);                                               
         if (s) {                                                                       
            InJis90 = TRUE;                                                             
         } /* endif */                                                                  
      } /* endif */                                                                     
   } /* endif */
   PULONG sv = new ULONG[QSV_MAX];
   rc = DosQuerySysInfo(1, QSV_MAX, sv, sizeof(ULONG) * QSV_MAX);
   if (rc != NO_ERROR) {
      return (rc);
   } /* endif */
   PCHAR fn = new CHAR[sv[QSV_MAX_PATH_LENGTH-1]+1];
   sprintf(fn, "%c:\\OS2\\SYSDATA\\$SYS1IBM.TBL", 'A' + sv[QSV_BOOT_DRIVE-1] - 1);
   delete sv;
   FILE* fp = fopen(fn, "rb");
   delete fn;
   PWCHAR_T *pccodetojiscode;
   PWCHAR_T *jiscodetopccode;
   if (InJis90) {
      pccodetojiscode = PccodeToJis83Table;
      jiscodetopccode = Jis90ToPccodeTable;
   } else {
      pccodetojiscode = PccodeToJis83Table;
      jiscodetopccode = Jis78ToPccodeTable;
   } /* endif */
   for (; ; ) {
      int h = fgetc(fp);
      if (EOF == h) {
         break;
      } /* endif */
      int l = fgetc(fp);
      if (EOF == l) {
         break;
      } /* endif */
      char pchi = h;
      char pclo = l;
      wchar_t pccode = h << ByteBitWidth | l;
      if (pccode == 0xfcfb || pccode == 0x2020) { /* skip invalid codes */
         continue;
      } /* endif */
      wchar_t jiscode = PcCodeToJisCode(pccode);
      SetValueInTable(pccodetojiscode, pclo, pchi, JisSpace, i, jiscode)
      h = jiscode >> ByteBitWidth;
      l = jiscode & ByteMask;
      SetValueInTable(jiscodetopccode, l, h, L'@', i, pccode)
   } /* endfor */
   fclose(fp);
   if (943 == DefaultCodepage) {
      for (int ku = 1; ku <= 119; ku += 1) {
         if (9 <= ku && ku <= 12 || 14 == ku || 15 == ku || 85 <= ku && ku <= 88 || 93 <= ku && ku <= 114) {
            continue;
         } /* endif */
         for (int ten = 1; ten <= 94; ten += 1) {
            wchar_t jiscode = KutenCodeToJisCode(100 * ku + ten);
            wchar_t pccode = JisCodeToPcCode(jiscode);
            int h = jiscode >> ByteBitWidth;
            int l = jiscode & ByteMask;
            SetValueInTable(jiscodetopccode, l, h, L'@', i, pccode)
         } /* endfor */
      } /* endfor */
   } /* endif */
   for (i = 0; i < sizeof(IbmCharTable) / sizeof(IbmCharTable[0]); i += 1) {
      int pclo = IbmCharTable[i].code1 & ByteMask;
      int pchi = IbmCharTable[i].code1 >> ByteBitWidth;
      wchar_t jiscode = PcCodeToJisCode(IbmCharTable[i].code2);
      SetValueInTable(pccodetojiscode, pclo, pchi, JisSpace, j, jiscode)
      if (IbmCharTable[i].code3) {
         jiscode = KutenCodeToJisCode(IbmCharTable[i].code3);
         int jislo = jiscode & ByteMask;
         int jishi = jiscode >> ByteBitWidth;
         SetValueInTable(JisX0212_1990ToPccodeTable, jislo, jishi, L'@', j, IbmCharTable[i].code1)
      } /* endif */
   } /* endfor */
   if (943 != DefaultCodepage) {
      for (i = 0; i < sizeof(NecCharTable) / sizeof(NecCharTable[0]); i += 1) {
         wchar_t jiscode = NecCharTable[i].code1;
         int jislo = jiscode & ByteMask;
         int jishi = jiscode >> ByteBitWidth;
         SetValueInTable(jiscodetopccode, jislo, jishi, JisSpace, j, NecCharTable[i].code2)
      } /* endfor */
      for (i = 89; i <= 92; i += 1) {
         int m = i == 92 ? 78 : 94;
         for (int j = 1; j <= m; j += 1) {
            int neckuten = 100 * i + j;
            int necjiscode = KutenCodeToJisCode(neckuten);
            int ibmkuten = (j + 28) > 94 ? (100 * (i + 27) + j + 28 - 94) : (100 * (i + 26) + j + 28);
            int pccode = JisCodeToPcCode(KutenCodeToJisCode(ibmkuten));
            int jislo = necjiscode & ByteMask;
            int jishi = necjiscode >> ByteBitWidth;
            SetValueInTable(jiscodetopccode, jislo, jishi, JisSpace, k, pccode)
         } /* endfor */
      } /* endfor */
   } /* endif */
   if (InJis90) {
      PccodeToJis83Table[0xfa][0xd0] = 0x3937; /* V */
      PccodeToJis83Table[0x8d][0x56] = 0x9452; /*  */
      PccodeToJis83Table[0xea][0xa4] = 0x5f66; /*  */
      PccodeToJis83Table[0xe0][0x86] = JisSpace;
      PccodeToJis83Table[0xea][0xa3] = JisSpace;
      memcpy(Jis83ToPccodeTable, Jis90ToPccodeTable, sizeof(PWCHAR_T) * (UCHAR_MAX + 1));
      Jis83ToPccodeTable[0x39] = new wchar_t[UCHAR_MAX + 1];
      memcpy(Jis83ToPccodeTable[0x39], Jis90ToPccodeTable[0x39], sizeof(wchar_t) * (UCHAR_MAX + 1));
      Jis90ToPccodeTable[0x39][0x37] = 0xfad0; /*  */
      Jis83ToPccodeTable[0x94] = new wchar_t[UCHAR_MAX + 1];
      memcpy(Jis83ToPccodeTable[0x94], Jis90ToPccodeTable[0x94], sizeof(wchar_t) * (UCHAR_MAX + 1));
      Jis90ToPccodeTable[0x94][0x52] = 0x8d56; /*  */
      Jis83ToPccodeTable[0x5f] = new wchar_t[UCHAR_MAX + 1];
      memcpy(Jis83ToPccodeTable[0x5f], Jis90ToPccodeTable[0x5f], sizeof(wchar_t) * (UCHAR_MAX + 1));
      Jis90ToPccodeTable[0x5f][0x66] = 0xeaa4; /*  */
      memcpy(Jis78ToPccodeTable, Jis83ToPccodeTable, sizeof(PWCHAR_T) * (UCHAR_MAX + 1));
      for (i = 0; i < NoOfJisEntries; i += 1) {
         if (!JisTable[i].oldjis.kutencode) { 
            continue;
         } /* endif */
         wchar_t jiscode = PcCodeToJisCode(JisTable[i].oldjis.pccode); 
         int l = jiscode & ByteMask;
         int h = jiscode >> ByteBitWidth;
         if (Jis83ToPccodeTable[h] == Jis78ToPccodeTable[h]) {
            Jis78ToPccodeTable[h] = new wchar_t[UCHAR_MAX + 1];
            memcpy(Jis78ToPccodeTable[h], Jis83ToPccodeTable[h], sizeof(wchar_t) * (UCHAR_MAX + 1));
         } /* endif */
         Jis83ToPccodeTable[h][l] = JisTable[i].newjis.pccode; 
      } /* endfor */
   } else {
      for (i = 0; i < NoOfJisEntries; i += 1) {
         wchar_t pccode = JisTable[i].oldjis.pccode; 
         int l = pccode & ByteMask;
         int h = pccode >> ByteBitWidth;
         wchar_t kutencode = JisTable[i].oldjis.kutencode; 
         if (kutencode) {
            PccodeToJis83Table[h][l] = KutenCodeToJisCode(kutencode);
         } else {
            PccodeToJis83Table[h][l] = L'@';
         } /* endif */
      } /* endfor */
      memcpy(Jis83ToPccodeTable, Jis78ToPccodeTable, sizeof(PWCHAR_T) * (UCHAR_MAX + 1));
      for (i = 0; i < NoOfJisEntries; i += 1) {
         wchar_t kutencode = JisTable[i].oldjis.kutencode; 
         if (!kutencode) {
            continue;
         } /* endif */
         int l = kutencode % 100 + 0x20;
         int h = kutencode / 100 + 0x20;
         if (Jis83ToPccodeTable[h] == Jis78ToPccodeTable[h]) {
            Jis83ToPccodeTable[h] = new wchar_t[UCHAR_MAX + 1];
            memcpy(Jis83ToPccodeTable[h], Jis78ToPccodeTable[h], sizeof(wchar_t) * (UCHAR_MAX + 1));
         } /* endif */
         Jis83ToPccodeTable[h][l] = JisTable[i].oldjis.pccode; 
      } /* endfor */
      memcpy(Jis90ToPccodeTable, Jis83ToPccodeTable, sizeof(PWCHAR_T) * (UCHAR_MAX + 1));
      Jis90ToPccodeTable[0x94] = new wchar_t[UCHAR_MAX + 1];
      memcpy(Jis90ToPccodeTable[0x94], Jis83ToPccodeTable[0x94], sizeof(wchar_t) * (UCHAR_MAX + 1));
      Jis90ToPccodeTable[0x94][0x52] = 0x8d56; /* V */
      Jis90ToPccodeTable[0x39] = new wchar_t[UCHAR_MAX + 1];
      memcpy(Jis90ToPccodeTable[0x39], Jis83ToPccodeTable[0x39], sizeof(wchar_t) * (UCHAR_MAX + 1));
      Jis90ToPccodeTable[0x39][0x37] = 0xfad0; /*  */
      Jis90ToPccodeTable[0x74] = new wchar_t[UCHAR_MAX + 1];
      memcpy(Jis90ToPccodeTable[0x74], Jis83ToPccodeTable[0x74], sizeof(wchar_t) * (UCHAR_MAX + 1));
      Jis90ToPccodeTable[0x74][0x26] = 0xe086; /*  */
      Jis90ToPccodeTable[0x74][0x25] = L'@'; /*  */
      Jis90ToPccodeTable[0x5f] = new wchar_t[UCHAR_MAX + 1];
      memcpy(Jis90ToPccodeTable[0x5f], Jis83ToPccodeTable[0x5f], sizeof(wchar_t) * (UCHAR_MAX + 1));
      Jis90ToPccodeTable[0x5f][0x66] =  L'@'; /*  */
   } /* endif */
   NoOfMimeEncodedWord = 2*NoOfMimeCharSet;
   MimeEncodedWordTable = new MimeCharSet[NoOfMimeEncodedWord];
   for (i = 0; i < NoOfMimeCharSet; i += 1) {
      int j = 2 * i;
      MimeEncodedWordTable[j].namesize = MimeEncodedWordTable[j+1].namesize = 2 + MimeCharSetTable[i].namesize + 3;
      MimeEncodedWordTable[j].charset = MimeEncodedWordTable[j+1].charset = MimeCharSetTable[i].charset;
      MimeEncodedWordTable[j].name = new char[MimeEncodedWordTable[j].namesize+1];
      MimeEncodedWordTable[j+1].name = new char[MimeEncodedWordTable[j+1].namesize+1];
      sprintf(MimeEncodedWordTable[j].name, "=?%s?B?", MimeCharSetTable[i].name);
      sprintf(MimeEncodedWordTable[j+1].name, "=?%s?Q?", MimeCharSetTable[i].name);
   } /* endfor */
   qsort(MimeEncodedWordTable, NoOfMimeEncodedWord, sizeof(MimeEncodedWordTable[0]), compcharset);
   UppercaseTable[0] = new wchar_t[UCHAR_MAX+1];
   ToSbcsTable[0] = new PCHAR[UCHAR_MAX+1];
   for (int j = 0; j <= UCHAR_MAX; j += 1) {
      UppercaseTable[0][j] = j;
      PCHAR t = new CHAR[2];
      ToSbcsTable[0][j] = NULL;
      ToDbcsTable1[j] = j;
   } /* endfor */
   for (i = 0; i < sizeof(JisCaseTable) / sizeof(JisCaseTable[0]); i += 1) {
      if (JisCaseTable[i].sbupper) {
         UppercaseTable[0][JisCaseTable[i].sb[0]] = JisCaseTable[i].sbupper[0];
      } /* endif */
      int h = JisCaseTable[i].db >> ByteBitWidth;
      int l = JisCaseTable[i].db & ByteMask;
      if (!UppercaseTable[h]) {
         UppercaseTable[h] = new wchar_t[UCHAR_MAX+1];
         int c = h << ByteBitWidth;
         for (j = 0; j <= UCHAR_MAX; j += 1) {
            UppercaseTable[h][j] = c | j;
         } /* endfor */
      } /* endif */
      if (!ToSbcsTable[h]) {
         ToSbcsTable[h] = new PCHAR[UCHAR_MAX+1];
         int c = h << ByteBitWidth;
         for (j = 0; j <= UCHAR_MAX; j += 1) {
            ToSbcsTable[h][j] = NULL;
         } /* endfor */
      } /* endif */
      UppercaseTable[h][l] = JisCaseTable[i].dbupper;
      ToSbcsTable[h][l] = JisCaseTable[i].sb;
      ToDbcsTable1[j] = JisCaseTable[i].db;
      if (JisCaseTable[i].sb) {
         if (strlen(JisCaseTable[i].sb) > 1) {
            if ('' == JisCaseTable[i].sb[1]) {
               ToDbcsTable3[JisCaseTable[i].sb[0]] = JisCaseTable[i].db;
            } else {
               ToDbcsTable2[JisCaseTable[i].sb[0]] = JisCaseTable[i].db;
            } /* endif */
         } else {
            ToDbcsTable1[JisCaseTable[i].sb[0]] = JisCaseTable[i].db;
         } /* endif */
      } /* endif */
   } /* endfor */
   for (i = 0; i < sizeof(JisDbcsConvTable1) / sizeof(JisDbcsConvTable1[0]); i += 1) {
      int h = JisDbcsConvTable1[i].code1 >> ByteBitWidth;
      int l = JisDbcsConvTable1[i].code1 & ByteMask;
      if (!ToDbcsTable4[h]) {
         ToDbcsTable4[h] = new wchar_t[UCHAR_MAX+1];
         int c = h << ByteBitWidth;
         for (j = 0; j <= UCHAR_MAX; j += 1) {
            ToDbcsTable4[h][j] = c | j;
         } /* endfor */
      } /* endif */
      ToDbcsTable4[h][l] = JisDbcsConvTable1[i].code2;
      h = JisDbcsConvTable1[i].code2 >> ByteBitWidth;
      l = JisDbcsConvTable1[i].code2 & ByteMask;
      if (!UppercaseTable[h]) {
         UppercaseTable[h] = new wchar_t[UCHAR_MAX+1];
         int c = h << ByteBitWidth;
         for (j = 0; j <= UCHAR_MAX; j += 1) {
            UppercaseTable[h][j] = c | j;
         } /* endfor */
      } /* endif */
      UppercaseTable[h][l] = JisDbcsConvTable1[i].code1;
   } /* endfor */
   for (i = 0; i < sizeof(JisDbcsConvTable2) / sizeof(JisDbcsConvTable2[0]); i += 1) {
      int h = JisDbcsConvTable2[i].code1 >> ByteBitWidth;
      int l = JisDbcsConvTable2[i].code1 & ByteMask;
      if (!ToDbcsTable5[h]) {
         ToDbcsTable5[h] = new wchar_t[UCHAR_MAX+1];
         int c = h << ByteBitWidth;
         for (j = 0; j <= UCHAR_MAX; j += 1) {
            ToDbcsTable5[h][j] = c | j;
         } /* endfor */
      } /* endif */
      ToDbcsTable5[h][l] = JisDbcsConvTable2[i].code2;
      h = JisDbcsConvTable2[i].code2 >> ByteBitWidth;
      l = JisDbcsConvTable2[i].code2 & ByteMask;
      if (!UppercaseTable[h]) {
         UppercaseTable[h] = new wchar_t[UCHAR_MAX+1];
         int c = h << ByteBitWidth;
         for (j = 0; j <= UCHAR_MAX; j += 1) {
            UppercaseTable[h][j] = c | j;
         } /* endfor */
      } /* endif */
      UppercaseTable[h][l] = JisDbcsConvTable2[i].code1;
   } /* endfor */
   memset(Cp850ToIso8859_1Table, ' ', UCHAR_MAX + 1);
   memset(Cp852ToIso8859_2Table, ' ', UCHAR_MAX + 1);
   memset(Cp857ToIso8859_9Table, ' ', UCHAR_MAX + 1);
   for (i = 0; i <= UCHAR_MAX; i += 1) {
      Cp850ToIso8859_1Table[Iso8859_1ToCp850Table[i]] = i;
      Cp852ToIso8859_2Table[Iso8859_2ToCp852Table[i]] = i;
      Cp857ToIso8859_9Table[Iso8859_9ToCp857Table[i]] = i;
   } /* endfor */
   CHAR name[BUFSIZ];
   HMODULE hmod;
   rc = DosLoadModule(name, sizeof(name), "PMNLS", &hmod);
   if (rc != NO_ERROR) {
      return (rc);
   } /* endif */
   rc = DosQueryProcAddr(hmod, 0L, "NLSFINDWORD", (PFN*)&NlsFindWord);
   if (rc != NO_ERROR) {
      return (rc);
   } /* endif */
   COUNTRYCODE cc;
   cc.country = 0;
   cc.codepage = DefaultCodepage;
   CHAR vec[BUFSIZ];
   rc = DosQueryDBCSEnv(sizeof(vec), &cc, vec);
   if (rc != NO_ERROR) {
      return (rc);
   } /* endif */
   for (i = 0; vec[i]; i += 2) {
      for (j = vec[i]; j <= vec[i+1]; j += 1) {
         DbcsVec[j] = TRUE;
      } /* endfor */
   } /* endfor */
   TableIsLoaded = TRUE;
   return (NO_ERROR);
}

ULONG _Export 
RxJisLoadFuncs(
PUCHAR name, 
ULONG argc,
RXSTRING argv[],
PSZ queuename,
RXSTRING *retstr) {
   if (argc > 0) {
      return (CallingInvalid);
   } /* endif */
   for (int i = 0; i < NoOfEntries; i += 1) {
      ULONG code = RexxDeregisterFunction(RxJisFuncsTable[i]);
      switch (code) {
      case RXFUNC_OK:
      case RXFUNC_DEFINED:
      case RXFUNC_NOTREG:
         break;
      default:
         return (CallingInvalid);
      } /* endswitch */
      code = RexxRegisterFunctionDll(RxJisFuncsTable[i], RxJisDll, RxJisFuncsTable[i]);
      switch (code) {
      case RXFUNC_OK:
         break;
      default:
         return (CallingInvalid);
      } /* endswitch */
      if (code != RXFUNC_OK && code != RXFUNC_DEFINED) {
         return (CallingInvalid);
      } /* endif */
   } /* endfor */
   if (LoadRxJisTables() != NO_ERROR) {
      return (CallingInvalid);
   } /* endif */
   i = sprintf(RXSTRPTR(*retstr), "RXJIS Version: %s (%s %s)", RxJisVersion, __DATE__, __TIME__);
   if (i + 1 > RXSTRLEN(*retstr)) {
      return (CallingInvalid);
   } /* endif */
   retstr->strlength = i;
   return (CallingValid);
}

ULONG _Export  
RxJisDropFuncs(
PUCHAR name, 
ULONG argc,
RXSTRING argv[],
PSZ queuename,
RXSTRING *retstr) {
   if (argc > 0) {
      return (CallingInvalid);
   } /* endif */
   for (int i = 0; i < NoOfEntries; i += 1) {
      ULONG code = RexxDeregisterFunction(RxJisFuncsTable[i]);
      switch (code) {
      case RXFUNC_OK:
      case RXFUNC_DEFINED:
      case RXFUNC_NOTREG:
         break;
      default:
         return (CallingInvalid);
      } /* endswitch */
   } /* endfor */
   retstr->strlength = 0;
   return (CallingValid);
}
